""" Test everything related to requests """
import json
from copy import deepcopy

from django.http import JsonResponse

from automated_logging.models import RequestEvent
from automated_logging.tests.base import BaseTestCase, USER_CREDENTIALS
from automated_logging.tests.helpers import random_string


class LoggedOutRequestsTestCase(BaseTestCase):
    def setUp(self):
        from django.conf import settings
        from automated_logging.settings import settings as conf

        super().setUp()

        settings.AUTOMATED_LOGGING['request']['exclude']['applications'] = []
        conf.load.cache_clear()

        RequestEvent.objects.all().delete()

    @staticmethod
    def view(request):
        return JsonResponse({})

    def test_simple(self):
        self.bypass_request_restrictions()

        self.request('GET', self.view)

        events = RequestEvent.objects.all()
        self.assertEqual(events.count(), 1)

        event = events[0]

        self.assertEqual(event.user, None)


class LoggedInRequestsTestCase(BaseTestCase):
    def setUp(self):
        from django.conf import settings
        from automated_logging.settings import settings as conf

        super().setUp()

        settings.AUTOMATED_LOGGING['request']['exclude']['applications'] = []
        conf.load.cache_clear()

        self.client.login(**USER_CREDENTIALS)

        RequestEvent.objects.all().delete()

    @staticmethod
    def view(request):
        return JsonResponse({})

    def test_simple(self):
        self.bypass_request_restrictions()

        self.request('GET', self.view)

        events = RequestEvent.objects.all()
        self.assertEqual(events.count(), 1)

        event = events[0]

        self.assertEqual(event.ip, '127.0.0.1')
        self.assertEqual(event.user, self.user)
        self.assertEqual(event.status, 200)
        self.assertEqual(event.method, 'GET')
        self.assertEqual(event.uri, '/')

    def test_404(self):
        self.bypass_request_restrictions()

        self.client.get(f'/{random_string()}')

        events = RequestEvent.objects.all()
        self.assertEqual(events.count(), 1)

        event = events[0]
        self.assertEqual(event.status, 404)

    @staticmethod
    def exception(request):
        raise Exception

    def test_500(self):
        self.bypass_request_restrictions()

        try:
            self.request('GET', self.exception)
        except:
            pass

        events = RequestEvent.objects.all()
        self.assertEqual(events.count(), 1)

        event = events[0]
        self.assertGreaterEqual(event.status, 500)


class DataRecordingRequestsTestCase(BaseTestCase):
    def setUp(self):
        from django.conf import settings
        from automated_logging.settings import settings as conf

        super().setUp()

        settings.AUTOMATED_LOGGING['request']['exclude']['applications'] = []
        settings.AUTOMATED_LOGGING['request']['data']['enabled'] = [
            'response',
            'request',
        ]
        conf.load.cache_clear()

        self.client.login(**USER_CREDENTIALS)

        RequestEvent.objects.all().delete()

    @staticmethod
    def view(request):
        return JsonResponse({'test': 'example'})

    def test_payload(self):
        # TODO: preliminary until request/response parsing is implemented
        from django.conf import settings
        from automated_logging.settings import settings as conf

        self.bypass_request_restrictions()

        settings.AUTOMATED_LOGGING['request']['data']['enabled'] = [
            'request',
            'response',
        ]
        conf.load.cache_clear()

        self.request('GET', self.view, data=json.dumps({'X': 'Y'}))

        events = RequestEvent.objects.all()
        self.assertEqual(events.count(), 1)

        response = json.dumps({'test': 'example'})
        request = json.dumps({'X': 'Y'})
        event = events[0]
        self.assertEqual(event.response.content.decode(), response)
        self.assertEqual(event.request.content.decode(), request)

    def test_exclusion_by_application(self):
        self.request('GET', self.view)
        self.assertEqual(RequestEvent.objects.count(), 0)



"""
Everything related to the admin interface of ModelMirror is located in here
"""
from django.contrib.admin import register
from django.shortcuts import redirect

from automated_logging.admin.base import ReadOnlyAdminMixin, ReadOnlyTabularInlineMixin
from automated_logging.models import ModelMirror, ModelField


class ModelFieldInline(ReadOnlyTabularInlineMixin):
    """ list all recorded fields """

    model = ModelField

    fields = ['name', 'type']

    verbose_name = 'Recorded Field'
    verbose_name_plural = 'Recorded Fields'


@register(ModelMirror)
class ModelMirrorAdmin(ReadOnlyAdminMixin):
    """ admin page specification for ModelMirror """

    def has_module_permission(self, request):
        """ prevents this from showing up index.html """
        return False

    def changelist_view(self, request, **kwargs):
        """ instead of showing the changelist view redirect to the parent app_list"""
        return redirect('admin:app_list', self.model._meta.app_label)

    date_hierarchy = 'updated_at'
    ordering = ('-updated_at',)

    fieldsets = (('Information', {'fields': ('id', 'application', 'name')},),)

    inlines = [ModelFieldInline]



# Generated by Django 3.0.7 on 2020-09-13 18:33

from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):
    from django.conf import settings

    dependencies = [
        ('automated_logging', '0017_auto_20200819_1004'),
    ]

    if hasattr(settings, 'AUTOMATED_LOGGING_DEV') and settings.AUTOMATED_LOGGING_DEV:
        operations = [
            migrations.CreateModel(
                name='DecoratorOverrideExclusionTest',
                fields=[
                    (
                        'id',
                        models.UUIDField(
                            default=uuid.uuid4, primary_key=True, serialize=False
                        ),
                    ),
                    ('created_at', models.DateTimeField(auto_now_add=True)),
                    ('updated_at', models.DateTimeField(auto_now=True)),
                    ('random', models.CharField(max_length=255, null=True)),
                    ('random2', models.CharField(max_length=255, null=True)),
                ],
            ),
            migrations.CreateModel(
                name='FullClassBasedExclusionTest',
                fields=[
                    (
                        'id',
                        models.UUIDField(
                            default=uuid.uuid4, primary_key=True, serialize=False
                        ),
                    ),
                    ('created_at', models.DateTimeField(auto_now_add=True)),
                    ('updated_at', models.DateTimeField(auto_now=True)),
                    ('random', models.CharField(max_length=255, null=True)),
                    ('random2', models.CharField(max_length=255, null=True)),
                ],
            ),
            migrations.CreateModel(
                name='FullDecoratorBasedExclusionTest',
                fields=[
                    (
                        'id',
                        models.UUIDField(
                            default=uuid.uuid4, primary_key=True, serialize=False
                        ),
                    ),
                    ('created_at', models.DateTimeField(auto_now_add=True)),
                    ('updated_at', models.DateTimeField(auto_now=True)),
                    ('random', models.CharField(max_length=255, null=True)),
                    ('random2', models.CharField(max_length=255, null=True)),
                ],
            ),
            migrations.CreateModel(
                name='OrdinaryTest',
                fields=[
                    (
                        'id',
                        models.UUIDField(
                            default=uuid.uuid4, primary_key=True, serialize=False
                        ),
                    ),
                    ('created_at', models.DateTimeField(auto_now_add=True)),
                    ('updated_at', models.DateTimeField(auto_now=True)),
                    ('random', models.CharField(max_length=255, null=True)),
                    ('random2', models.CharField(max_length=255, null=True)),
                ],
            ),
            migrations.CreateModel(
                name='PartialClassBasedExclusionTest',
                fields=[
                    (
                        'id',
                        models.UUIDField(
                            default=uuid.uuid4, primary_key=True, serialize=False
                        ),
                    ),
                    ('created_at', models.DateTimeField(auto_now_add=True)),
                    ('updated_at', models.DateTimeField(auto_now=True)),
                    ('random', models.CharField(max_length=255, null=True)),
                    ('random2', models.CharField(max_length=255, null=True)),
                ],
            ),
            migrations.CreateModel(
                name='PartialDecoratorBasedExclusionTest',
                fields=[
                    (
                        'id',
                        models.UUIDField(
                            default=uuid.uuid4, primary_key=True, serialize=False
                        ),
                    ),
                    ('created_at', models.DateTimeField(auto_now_add=True)),
                    ('updated_at', models.DateTimeField(auto_now=True)),
                    ('random', models.CharField(max_length=255, null=True)),
                    ('random2', models.CharField(max_length=255, null=True)),
                ],
            ),
            migrations.CreateModel(
                name='SpeedTest',
                fields=[
                    (
                        'id',
                        models.UUIDField(
                            default=uuid.uuid4, primary_key=True, serialize=False
                        ),
                    ),
                    ('created_at', models.DateTimeField(auto_now_add=True)),
                    ('updated_at', models.DateTimeField(auto_now=True)),
                    ('column0', models.CharField(max_length=15, null=True)),
                    ('column1', models.CharField(max_length=15, null=True)),
                    ('column2', models.CharField(max_length=15, null=True)),
                    ('column3', models.CharField(max_length=15, null=True)),
                    ('column4', models.CharField(max_length=15, null=True)),
                    ('column5', models.CharField(max_length=15, null=True)),
                    ('column6', models.CharField(max_length=15, null=True)),
                    ('column7', models.CharField(max_length=15, null=True)),
                    ('column8', models.CharField(max_length=15, null=True)),
                    ('column9', models.CharField(max_length=15, null=True)),
                    ('column10', models.CharField(max_length=15, null=True)),
                    ('column11', models.CharField(max_length=15, null=True)),
                    ('column12', models.CharField(max_length=15, null=True)),
                    ('column13', models.CharField(max_length=15, null=True)),
                    ('column14', models.CharField(max_length=15, null=True)),
                    ('column15', models.CharField(max_length=15, null=True)),
                    ('column16', models.CharField(max_length=15, null=True)),
                    ('column17', models.CharField(max_length=15, null=True)),
                    ('column18', models.CharField(max_length=15, null=True)),
                    ('column19', models.CharField(max_length=15, null=True)),
                    ('column20', models.CharField(max_length=15, null=True)),
                    ('column21', models.CharField(max_length=15, null=True)),
                    ('column22', models.CharField(max_length=15, null=True)),
                    ('column23', models.CharField(max_length=15, null=True)),
                    ('column24', models.CharField(max_length=15, null=True)),
                    ('column25', models.CharField(max_length=15, null=True)),
                    ('column26', models.CharField(max_length=15, null=True)),
                    ('column27', models.CharField(max_length=15, null=True)),
                    ('column28', models.CharField(max_length=15, null=True)),
                    ('column29', models.CharField(max_length=15, null=True)),
                    ('column30', models.CharField(max_length=15, null=True)),
                    ('column31', models.CharField(max_length=15, null=True)),
                    ('column32', models.CharField(max_length=15, null=True)),
                    ('column33', models.CharField(max_length=15, null=True)),
                    ('column34', models.CharField(max_length=15, null=True)),
                    ('column35', models.CharField(max_length=15, null=True)),
                    ('column36', models.CharField(max_length=15, null=True)),
                    ('column37', models.CharField(max_length=15, null=True)),
                    ('column38', models.CharField(max_length=15, null=True)),
                    ('column39', models.CharField(max_length=15, null=True)),
                    ('column40', models.CharField(max_length=15, null=True)),
                    ('column41', models.CharField(max_length=15, null=True)),
                    ('column42', models.CharField(max_length=15, null=True)),
                    ('column43', models.CharField(max_length=15, null=True)),
                    ('column44', models.CharField(max_length=15, null=True)),
                    ('column45', models.CharField(max_length=15, null=True)),
                    ('column46', models.CharField(max_length=15, null=True)),
                    ('column47', models.CharField(max_length=15, null=True)),
                    ('column48', models.CharField(max_length=15, null=True)),
                    ('column49', models.CharField(max_length=15, null=True)),
                    ('column50', models.CharField(max_length=15, null=True)),
                    ('column51', models.CharField(max_length=15, null=True)),
                    ('column52', models.CharField(max_length=15, null=True)),
                    ('column53', models.CharField(max_length=15, null=True)),
                    ('column54', models.CharField(max_length=15, null=True)),
                    ('column55', models.CharField(max_length=15, null=True)),
                    ('column56', models.CharField(max_length=15, null=True)),
                    ('column57', models.CharField(max_length=15, null=True)),
                    ('column58', models.CharField(max_length=15, null=True)),
                    ('column59', models.CharField(max_length=15, null=True)),
                    ('column60', models.CharField(max_length=15, null=True)),
                    ('column61', models.CharField(max_length=15, null=True)),
                    ('column62', models.CharField(max_length=15, null=True)),
                    ('column63', models.CharField(max_length=15, null=True)),
                    ('column64', models.CharField(max_length=15, null=True)),
                    ('column65', models.CharField(max_length=15, null=True)),
                    ('column66', models.CharField(max_length=15, null=True)),
                    ('column67', models.CharField(max_length=15, null=True)),
                    ('column68', models.CharField(max_length=15, null=True)),
                    ('column69', models.CharField(max_length=15, null=True)),
                    ('column70', models.CharField(max_length=15, null=True)),
                    ('column71', models.CharField(max_length=15, null=True)),
                    ('column72', models.CharField(max_length=15, null=True)),
                    ('column73', models.CharField(max_length=15, null=True)),
                    ('column74', models.CharField(max_length=15, null=True)),
                    ('column75', models.CharField(max_length=15, null=True)),
                    ('column76', models.CharField(max_length=15, null=True)),
                    ('column77', models.CharField(max_length=15, null=True)),
                    ('column78', models.CharField(max_length=15, null=True)),
                    ('column79', models.CharField(max_length=15, null=True)),
                    ('column80', models.CharField(max_length=15, null=True)),
                    ('column81', models.CharField(max_length=15, null=True)),
                    ('column82', models.CharField(max_length=15, null=True)),
                    ('column83', models.CharField(max_length=15, null=True)),
                    ('column84', models.CharField(max_length=15, null=True)),
                    ('column85', models.CharField(max_length=15, null=True)),
                    ('column86', models.CharField(max_length=15, null=True)),
                    ('column87', models.CharField(max_length=15, null=True)),
                    ('column88', models.CharField(max_length=15, null=True)),
                    ('column89', models.CharField(max_length=15, null=True)),
                    ('column90', models.CharField(max_length=15, null=True)),
                    ('column91', models.CharField(max_length=15, null=True)),
                    ('column92', models.CharField(max_length=15, null=True)),
                    ('column93', models.CharField(max_length=15, null=True)),
                    ('column94', models.CharField(max_length=15, null=True)),
                    ('column95', models.CharField(max_length=15, null=True)),
                    ('column96', models.CharField(max_length=15, null=True)),
                    ('column97', models.CharField(max_length=15, null=True)),
                    ('column98', models.CharField(max_length=15, null=True)),
                    ('column99', models.CharField(max_length=15, null=True)),
                ],
            ),
            migrations.CreateModel(
                name='OneToOneTest',
                fields=[
                    (
                        'id',
                        models.UUIDField(
                            default=uuid.uuid4, primary_key=True, serialize=False
                        ),
                    ),
                    ('created_at', models.DateTimeField(auto_now_add=True)),
                    ('updated_at', models.DateTimeField(auto_now=True)),
                    (
                        'relationship',
                        models.OneToOneField(
                            null=True,
                            on_delete=django.db.models.deletion.CASCADE,
                            to='automated_logging.OrdinaryTest',
                        ),
                    ),
                ],
            ),
            migrations.CreateModel(
                name='M2MTest',
                fields=[
                    (
                        'id',
                        models.UUIDField(
                            default=uuid.uuid4, primary_key=True, serialize=False
                        ),
                    ),
                    ('created_at', models.DateTimeField(auto_now_add=True)),
                    ('updated_at', models.DateTimeField(auto_now=True)),
                    (
                        'relationship',
                        models.ManyToManyField(to='automated_logging.OrdinaryTest'),
                    ),
                ],
            ),
            migrations.CreateModel(
                name='ForeignKeyTest',
                fields=[
                    (
                        'id',
                        models.UUIDField(
                            default=uuid.uuid4, primary_key=True, serialize=False
                        ),
                    ),
                    ('created_at', models.DateTimeField(auto_now_add=True)),
                    ('updated_at', models.DateTimeField(auto_now=True)),
                    (
                        'relationship',
                        models.ForeignKey(
                            null=True,
                            on_delete=django.db.models.deletion.CASCADE,
                            to='automated_logging.OrdinaryTest',
                        ),
                    ),
                ],
            ),
        ]

    else:
        operations = []



"""
Everything related to the admin interface of UnspecifiedEvent is located in here
"""
from django.contrib.admin import register, RelatedOnlyFieldListFilter

from automated_logging.admin.base import ReadOnlyAdminMixin, ReadOnlyTabularInlineMixin
from automated_logging.models import UnspecifiedEvent


@register(UnspecifiedEvent)
class UnspecifiedEventAdmin(ReadOnlyAdminMixin):
    """ admin page specification for the UnspecifiedEvent """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_id(self, instance):
        """ shorten the id to the first 8 digits """
        return str(instance.id).split('-')[0]

    get_id.short_description = 'UUID'

    list_display = ('get_id', 'updated_at', 'level', 'message')

    date_hierarchy = 'updated_at'
    ordering = ('-updated_at',)

    list_filter = ('updated_at',)

    fieldsets = (
        (
            'Information',
            {'fields': ('id', 'updated_at', 'application', 'level', 'message')},
        ),
        ('Location', {'fields': ('file', 'line')}),
    )



# Generated by Django 2.1.1 on 2018-12-29 23:23

from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('automated_logging', '0014_auto_20180219_0859'),
    ]

    operations = [
        migrations.DeleteModel(
            name='LDAP',
        ),
        migrations.AlterModelOptions(
            name='unspecified',
            options={'verbose_name': 'Non DAL Message', 'verbose_name_plural': 'Non DAL Messages'},
        ),
        migrations.AlterField(
            model_name='application',
            name='id',
            field=models.UUIDField(db_index=True, default=uuid.uuid4, primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='field',
            name='id',
            field=models.UUIDField(db_index=True, default=uuid.uuid4, primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='model',
            name='id',
            field=models.UUIDField(db_index=True, default=uuid.uuid4, primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='modelchangelog',
            name='id',
            field=models.UUIDField(db_index=True, default=uuid.uuid4, primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='modelmodification',
            name='id',
            field=models.UUIDField(db_index=True, default=uuid.uuid4, primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='modelobject',
            name='id',
            field=models.UUIDField(db_index=True, default=uuid.uuid4, primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='request',
            name='id',
            field=models.UUIDField(db_index=True, default=uuid.uuid4, primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='unspecified',
            name='id',
            field=models.UUIDField(db_index=True, default=uuid.uuid4, primary_key=True, serialize=False),
        ),
    ]



# Generated by Django 3.0.7 on 2020-08-03 19:17
# Edited by Bilal Mahmoud for 5.x.x to 6.x.x conversion

from django.db import migrations
from django.conf import settings
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import picklefield.fields
import uuid

import logging

logger = logging.getLogger(__name__)


def convert(apps, schema_editor):
    """ convert from 5.x.x to 6.x.x """
    alias = schema_editor.connection.alias

    # convert all new applications
    logger.info('Converting Applications from 5.x.x to 6.x.x')
    # Application does not change, except new unknown Application
    Application = apps.get_model('automated_logging', 'Application')
    applications = [Application(name=None)]
    no_application = applications[0]

    # convert request events
    logger.info('Converting Request Events from 5.x.x to 6.x.x')
    RequestOld = apps.get_model('automated_logging', 'Request')
    RequestEvent = apps.get_model('automated_logging', 'RequestEvent')
    requests = [
        RequestEvent(
            id=r.id,
            created_at=r.created_at,
            updated_at=r.updated_at,
            user=r.user,
            uri=str(r.uri),
            method=r.method[:32],
            status=r.status,
            application=Application.objects.using(alias).get(id=r.application.id)
            if r.application
            else no_application,
        )
        for r in RequestOld.objects.using(alias).all()
    ]

    # convert unspecified events
    logger.info('Converting Unspecified Events from 5.x.x to 6.x.x')
    UnspecifiedOld = apps.get_model('automated_logging', 'Unspecified')
    UnspecifiedEvent = apps.get_model('automated_logging', 'UnspecifiedEvent')
    unspecified = [
        UnspecifiedEvent(
            id=u.id,
            created_at=u.created_at,
            updated_at=u.updated_at,
            message=u.message,
            level=u.level,
            file=u.file,
            line=u.line,
            application=no_application,
        )
        for u in UnspecifiedOld.objects.using(alias).all()
    ]

    # convert model events
    logger.info(
        'Converting Models Events from 5.x.x to 6.x.x (This might take a while)'
    )
    ModelOld = apps.get_model('automated_logging', 'Model')
    ModelEvent = apps.get_model('automated_logging', 'ModelEvent')

    ModelMirror = apps.get_model('automated_logging', 'ModelMirror')
    ModelEntry = apps.get_model('automated_logging', 'ModelEntry')
    ModelField = apps.get_model('automated_logging', 'ModelField')
    ModelRelationshipModification = apps.get_model(
        'automated_logging', 'ModelRelationshipModification'
    )
    ModelValueModification = apps.get_model(
        'automated_logging', 'ModelValueModification'
    )

    mirrors = [ModelMirror(name='', application=no_application)]
    no_mirror = mirrors[0]

    entries = [ModelEntry(value='', primary_key='', mirror=no_mirror)]
    no_entry = entries[0]

    fields = [ModelField(name='', mirror=no_mirror, type='')]
    no_field = fields[0]

    events = []

    value_modifications = []
    relationship_modifications = []

    def get_application(content_type):
        """
        simple helper function to get a new application from a content type
        :return: ApplicationNew
        """
        app = next(
            (a for a in applications if a.name == content_type.app_label),
            None,
        )

        if not app:
            app = Application(name=content_type.app_label)
            applications.append(app)

        return app

    def get_mirror(content_type):
        """
        simple helper function to get the new mirror from content type
        :return: ModelMirror
        """
        app = get_application(content_type)
        mir = next((m for m in mirrors if m.name == content_type.model), None)
        if not mir:
            mir = ModelMirror(name=content_type.model, application=app)
            mirrors.append(mir)

        return mir

    def get_entry(value, mir):
        """
        simple helper function to find the appropriate entry
        :return: ModelEntry
        """
        ent = next((e for e in entries if e.value == value), None)
        if not ent:
            ent = ModelEntry(value=value, primary_key='', mirror=mir)
            entries.append(ent)
        return ent

    def get_field(target):
        """
        simple inline helper function to get the correct field
        :return: ModelField
        """

        if target is None:
            return no_field

        app = get_application(target.model)

        mir = next(
            (m for m in mirrors if m.application.name == target.model.app_label),
            None,
        )
        if not mir:
            mir = ModelMirror(name=target.model.model, application=app)

        fie = next(
            (e for e in fields if e.name == target.name and e.mirror == mir),
            None,
        )
        if not fie:
            fie = ModelField(name=target.name, mirror=mir, type='<UNKNOWN>')
            fields.append(fie)

        return fie

    oldies = ModelOld.objects.using(alias).all()

    progress = oldies.count() // 10
    idx = 0
    for old in oldies:
        event = ModelEvent(
            id=old.id, created_at=old.created_at, updated_at=old.updated_at
        )

        # converting old.information
        # old.information -> event.model
        #   value => value
        #   content_type => ModelMirror
        #       app_label => Application
        #       model => ModelMirror.name
        # provide some defaults, so that if no content_type is there
        # the program doesn't poo all over the floor.
        entry = no_entry
        if old.information:
            mirror = no_mirror

            # there is a chance old.information.type is null,
            # we need to check for that case
            if old.information.type:
                mirror = get_mirror(old.information.type)

            entry = next((e for e in entries if e.value == old.information.value), None)
            if not entry:
                entry = ModelEntry(
                    value=old.information.value, primary_key='', mirror=mirror
                )
                entries.append(entry)
        event.entry = entry

        ACTION_TRANSLATIONS = {0: None, 1: 1, 2: 0, 3: -1}

        # old.modification -> event.modifications, event.relationships
        # if information has content_type then relationship, else modification
        # problem: field is unknown
        #   modification => modifications
        #       operation.YOUCANDECIDE => operation
        #       previously       => previous
        #       currently        => current
        #   inserted => relationship
        #   removed => relationship

        event.operation = ACTION_TRANSLATIONS[old.action]

        if old.modification:
            for inserted in old.modification.inserted.all():
                mirror = get_mirror(inserted.type)
                rel = ModelRelationshipModification(
                    field=get_field(inserted.field),
                    entry=get_entry(inserted.value, mirror),
                    operation=1,
                )
                rel.event = event
                relationship_modifications.append(rel)

            for removed in old.modification.removed.all():
                mirror = get_mirror(removed.type)
                rel = ModelRelationshipModification(
                    field=get_field(removed.field),
                    entry=get_entry(removed.value, mirror),
                    operation=-1,
                )
                rel.event = event
                relationship_modifications.append(rel)

            if old.modification.modification:
                previously = {
                    f.field.id: f
                    for f in old.modification.modification.previously.all()
                }
                for currently in old.modification.modification.currently.all():
                    # skip None string values
                    if currently.value == 'None':
                        currently.value = None

                    operation = 0
                    previous = None
                    if currently.field.id not in previously:
                        operation = 1
                    else:
                        previous = previously[currently.field.id].value
                        # previous can be "None" string
                        if previous == 'None':
                            operation = 1
                            previous = None
                        if previous is not None and currently.value is None:
                            operation = -1

                    val = ModelValueModification(
                        operation=operation,
                        field=get_field(currently.field),
                        previous=previous,
                        current=currently.value,
                    )
                    val.event = event
                    value_modifications.append(val)

                for removed in set(
                    p.field.id for p in old.modification.modification.previously.all()
                ).difference(
                    c.field.id for c in old.modification.modification.currently.all()
                ):
                    removed = previously[removed]
                    # skip None string values
                    if removed.value == 'None':
                        continue

                    val = ModelValueModification(
                        operation=-1,
                        field=get_field(removed.field),
                        previous=removed.value,
                        current=None,
                    )
                    val.event = event
                    value_modifications.append(val)

        event.user = old.user
        events.append(event)
        if idx % progress == 0:
            logger.info(f'{(idx // progress) * 10}%...')
        idx += 1

    logger.info('Bulk Saving Converted Objects (This can take a while)')
    Application.objects.using(alias).bulk_create(applications)
    RequestEvent.objects.using(alias).bulk_create(requests)
    UnspecifiedEvent.objects.using(alias).bulk_create(unspecified)
    logger.info('Saved Application, RequestEvent and UnspecifiedEvent')

    ModelMirror.objects.using(alias).bulk_create(mirrors)
    ModelEntry.objects.using(alias).bulk_create(entries)
    ModelField.objects.using(alias).bulk_create(fields)
    logger.info('Saved ModelMirror, ModelEntry, ModelField')

    ModelEvent.objects.using(alias).bulk_create(events)

    ModelValueModification.objects.using(alias).bulk_create(value_modifications)
    ModelRelationshipModification.objects.using(alias).bulk_create(
        relationship_modifications
    )

    logger.info(
        'Saved ModelValueModification, ModelRelationshipModification and ModelEvent'
    )


class Migration(migrations.Migration):
    dependencies = [
        ('automated_logging', '0015_auto_20181229_2323'),
    ]

    operations = [
        # migrations.CreateModel(
        #     # create temporary model
        #     # will be renamed
        #     name='ApplicationTemp',
        #     fields=[
        #         (
        #             'id',
        #             models.UUIDField(
        #                 db_index=True,
        #                 default=uuid.uuid4,
        #                 primary_key=True,
        #                 serialize=False,
        #             ),
        #         ),
        #         ('created_at', models.DateTimeField(auto_now_add=True)),
        #         ('updated_at', models.DateTimeField(auto_now=True)),
        #         ('name', models.CharField(max_length=255, null=True)),
        #     ],
        #     options={
        #         'verbose_name': 'Application',
        #         'verbose_name_plural': 'Applications',
        #     },
        # ),
        migrations.CreateModel(
            name='ModelEntry',
            fields=[
                (
                    'id',
                    models.UUIDField(
                        db_index=True,
                        default=uuid.uuid4,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('value', models.TextField()),
                ('primary_key', models.TextField()),
            ],
            options={
                'verbose_name': 'Model Entry',
                'verbose_name_plural': 'Model Entries',
            },
        ),
        migrations.CreateModel(
            name='ModelEvent',
            fields=[
                (
                    'id',
                    models.UUIDField(
                        db_index=True,
                        default=uuid.uuid4,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                (
                    'operation',
                    models.SmallIntegerField(
                        choices=[(1, 'create'), (0, 'modify'), (-1, 'delete')],
                        null=True,
                        validators=[
                            django.core.validators.MinValueValidator(-1),
                            django.core.validators.MaxValueValidator(1),
                        ],
                    ),
                ),
                (
                    'snapshot',
                    picklefield.fields.PickledObjectField(editable=False, null=True),
                ),
                ('performance', models.DurationField(null=True)),
                (
                    'entry',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to='automated_logging.ModelEntry',
                    ),
                ),
                (
                    'user',
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                'verbose_name': 'Model Event',
                'verbose_name_plural': 'Model Events',
            },
        ),
        migrations.CreateModel(
            name='ModelField',
            fields=[
                (
                    'id',
                    models.UUIDField(
                        db_index=True,
                        default=uuid.uuid4,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=255)),
                ('type', models.CharField(max_length=255)),
            ],
            options={
                'verbose_name': 'Model Field',
                'verbose_name_plural': 'Model Fields',
            },
        ),
        migrations.CreateModel(
            name='RequestContext',
            fields=[
                (
                    'id',
                    models.UUIDField(
                        db_index=True,
                        default=uuid.uuid4,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                (
                    'content',
                    picklefield.fields.PickledObjectField(editable=False, null=True),
                ),
                ('type', models.CharField(max_length=255)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='UnspecifiedEvent',
            fields=[
                (
                    'id',
                    models.UUIDField(
                        db_index=True,
                        default=uuid.uuid4,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('message', models.TextField(null=True)),
                ('level', models.PositiveIntegerField(default=20)),
                ('line', models.PositiveIntegerField(null=True)),
                ('file', models.TextField(null=True)),
                (
                    'application',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to='automated_logging.Application',
                    ),
                ),
            ],
            options={
                'verbose_name': 'Unspecified Event',
                'verbose_name_plural': 'Unspecified Events',
            },
        ),
        migrations.CreateModel(
            name='RequestEvent',
            fields=[
                (
                    'id',
                    models.UUIDField(
                        db_index=True,
                        default=uuid.uuid4,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('uri', models.TextField()),
                ('status', models.PositiveSmallIntegerField()),
                ('method', models.CharField(max_length=32)),
                ('ip', models.GenericIPAddressField(null=True)),
                (
                    'application',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to='automated_logging.Application',
                    ),
                ),
                (
                    'request',
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='request_context',
                        to='automated_logging.RequestContext',
                    ),
                ),
                (
                    'response',
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='response_context',
                        to='automated_logging.RequestContext',
                    ),
                ),
                (
                    'user',
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                'verbose_name': 'Request Event',
                'verbose_name_plural': 'Request Events',
            },
        ),
        migrations.CreateModel(
            name='ModelValueModification',
            fields=[
                (
                    'id',
                    models.UUIDField(
                        db_index=True,
                        default=uuid.uuid4,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                (
                    'operation',
                    models.SmallIntegerField(
                        choices=[(1, 'create'), (0, 'modify'), (-1, 'delete')],
                        null=True,
                        validators=[
                            django.core.validators.MinValueValidator(-1),
                            django.core.validators.MaxValueValidator(1),
                        ],
                    ),
                ),
                ('previous', models.TextField(null=True)),
                ('current', models.TextField(null=True)),
                (
                    'event',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='modifications',
                        to='automated_logging.ModelEvent',
                    ),
                ),
                (
                    'field',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to='automated_logging.ModelField',
                    ),
                ),
            ],
            options={
                'verbose_name': 'Model Entry Event Value Modification',
                'verbose_name_plural': 'Model Entry Event Value Modifications',
            },
        ),
        migrations.CreateModel(
            name='ModelRelationshipModification',
            fields=[
                (
                    'id',
                    models.UUIDField(
                        db_index=True,
                        default=uuid.uuid4,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                (
                    'operation',
                    models.SmallIntegerField(
                        choices=[(1, 'create'), (0, 'modify'), (-1, 'delete')],
                        null=True,
                        validators=[
                            django.core.validators.MinValueValidator(-1),
                            django.core.validators.MaxValueValidator(1),
                        ],
                    ),
                ),
                (
                    'event',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='relationships',
                        to='automated_logging.ModelEvent',
                    ),
                ),
                (
                    'field',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to='automated_logging.ModelField',
                    ),
                ),
                (
                    'entry',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to='automated_logging.ModelEntry',
                    ),
                ),
            ],
            options={
                'verbose_name': 'Model Entry Event Relationship Modification',
                'verbose_name_plural': 'Model Entry Event Relationship Modifications',
            },
        ),
        migrations.CreateModel(
            name='ModelMirror',
            fields=[
                (
                    'id',
                    models.UUIDField(
                        db_index=True,
                        default=uuid.uuid4,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=255)),
                (
                    'application',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to='automated_logging.Application',
                    ),
                ),
            ],
            options={
                'verbose_name': 'Model Mirror',
                'verbose_name_plural': 'Model Mirrors',
            },
        ),
        migrations.AddField(
            model_name='modelfield',
            name='mirror',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to='automated_logging.ModelMirror',
            ),
        ),
        migrations.AddField(
            model_name='modelentry',
            name='mirror',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to='automated_logging.ModelMirror',
            ),
        ),
        migrations.AlterField(
            model_name='application',
            name='name',
            field=models.CharField(max_length=255, null=True),
        ),
        # end of adding
        # conversion
        migrations.RunPython(convert),
        # migrations.RenameModel('ApplicationTemp', 'Application'),
        # removing 5.x.x
        migrations.RemoveField(
            model_name='field',
            name='model',
        ),
        migrations.RemoveField(
            model_name='model',
            name='application',
        ),
        migrations.RemoveField(
            model_name='model',
            name='information',
        ),
        migrations.RemoveField(
            model_name='model',
            name='modification',
        ),
        migrations.RemoveField(
            model_name='model',
            name='user',
        ),
        migrations.RemoveField(
            model_name='modelchangelog',
            name='information',
        ),
        migrations.RemoveField(
            model_name='modelchangelog',
            name='inserted',
        ),
        migrations.RemoveField(
            model_name='modelchangelog',
            name='modification',
        ),
        migrations.RemoveField(
            model_name='modelchangelog',
            name='removed',
        ),
        migrations.RemoveField(
            model_name='modelmodification',
            name='currently',
        ),
        migrations.RemoveField(
            model_name='modelmodification',
            name='previously',
        ),
        migrations.RemoveField(
            model_name='modelobject',
            name='field',
        ),
        migrations.RemoveField(
            model_name='modelobject',
            name='type',
        ),
        migrations.RemoveField(
            model_name='request',
            name='application',
        ),
        migrations.RemoveField(
            model_name='request',
            name='user',
        ),
        migrations.DeleteModel(
            name='Unspecified',
        ),
        # migrations.DeleteModel(name='Application',),
        migrations.DeleteModel(
            name='Field',
        ),
        migrations.DeleteModel(
            name='Model',
        ),
        migrations.DeleteModel(
            name='ModelChangelog',
        ),
        migrations.DeleteModel(
            name='ModelModification',
        ),
        migrations.DeleteModel(
            name='ModelObject',
        ),
        migrations.DeleteModel(
            name='Request',
        ),
        # end of removing
    ]



""" Test all Many-To-Many related things """


import random

from automated_logging.helpers import Operation
from automated_logging.models import ModelEvent
from automated_logging.tests.models import (
    M2MTest,
    OrdinaryTest,
    OneToOneTest,
    ForeignKeyTest,
)
from automated_logging.signals.m2m import find_m2m_rel
from automated_logging.tests.base import BaseTestCase
from automated_logging.tests.helpers import random_string


class LoggedOutM2MRelationshipsTestCase(BaseTestCase):
    def setUp(self):
        super().setUp()

        # delete all previous model events
        ModelEvent.objects.all().delete()

    @staticmethod
    def generate_children(samples=10):
        """ generate X children that are going to be used in various tests """
        children = [OrdinaryTest(random=random_string()) for _ in range(samples)]
        [c.save() for c in children]

        return children

    def test_add(self):
        """ check if adding X members works correctly """

        samples = 10
        children = self.generate_children(samples)

        m2m = M2MTest()
        m2m.save()

        ModelEvent.objects.all().delete()
        m2m.relationship.add(*children)
        m2m.save()

        events = ModelEvent.objects.all()
        self.assertEqual(events.count(), 1)

        event = events[0]
        self.assertEqual(event.modifications.count(), 0)
        self.assertEqual(event.relationships.count(), samples)

        children = {str(c.id): c for c in children}

        for relationship in event.relationships.all():
            self.assertEqual(relationship.operation, int(Operation.CREATE))
            self.assertEqual(relationship.field.name, 'relationship')
            self.assertIn(relationship.entry.primary_key, children)

    def test_delete(self):
        """ check if deleting X elements works correctly """

        samples = 10
        removed = 5
        children = self.generate_children(samples)

        m2m = M2MTest()
        m2m.save()
        m2m.relationship.add(*children)
        m2m.save()
        ModelEvent.objects.all().delete()

        selected = random.sample(children, k=removed)
        m2m.relationship.remove(*selected)
        m2m.save()

        events = ModelEvent.objects.all()
        self.assertEqual(events.count(), 1)

        event = events[0]
        self.assertEqual(event.modifications.count(), 0)
        self.assertEqual(event.relationships.count(), removed)

        children = {str(c.id): c for c in children}
        for relationship in event.relationships.all():
            self.assertEqual(relationship.operation, int(Operation.DELETE))
            self.assertEqual(relationship.field.name, 'relationship')
            self.assertIn(relationship.entry.primary_key, children)

    def test_clear(self):
        """ test if clearing all elements works correctly """

        samples = 10
        children = self.generate_children(samples)

        m2m = M2MTest()
        m2m.save()
        m2m.relationship.add(*children)
        m2m.save()
        ModelEvent.objects.all().delete()

        m2m.relationship.clear()
        m2m.save()

        events = ModelEvent.objects.all()
        self.assertEqual(events.count(), 1)

        event = events[0]
        self.assertEqual(event.modifications.count(), 0)
        self.assertEqual(event.relationships.count(), samples)

        children = {str(c.id): c for c in children}
        for relationship in event.relationships.all():
            self.assertEqual(relationship.operation, int(Operation.DELETE))
            self.assertEqual(relationship.field.name, 'relationship')
            self.assertIn(relationship.entry.primary_key, children)

    def test_one2one(self):
        """
        test if OneToOne are correctly recognized,
        should be handled by save.py
        """

        o2o = OneToOneTest()
        o2o.save()

        subject = OrdinaryTest(random=random_string())
        subject.save()
        ModelEvent.objects.all().delete()

        o2o.relationship = subject
        o2o.save()

        events = ModelEvent.objects.all()
        self.assertEqual(events.count(), 1)

        event = events[0]

        self.assertEqual(event.modifications.count(), 1)
        self.assertEqual(event.relationships.count(), 0)

        modification = event.modifications.all()[0]
        self.assertEqual(modification.field.name, 'relationship_id')
        self.assertEqual(modification.current, repr(subject.pk))

    def test_foreign(self):
        """
        test if ForeignKey are correctly recognized.

        should be handled by save.py
        """

        fk = ForeignKeyTest()
        fk.save()

        subject = OrdinaryTest(random=random_string())
        subject.save()

        ModelEvent.objects.all().delete()

        fk.relationship = subject
        fk.save()

        events = ModelEvent.objects.all()
        self.assertEqual(events.count(), 1)

        event = events[0]

        self.assertEqual(event.modifications.count(), 1)
        self.assertEqual(event.relationships.count(), 0)

        modification = event.modifications.all()[0]
        self.assertEqual(modification.field.name, 'relationship_id')
        self.assertEqual(modification.current, repr(subject.pk))

    # def test_no_change(self):
    #     samples = 10
    #     children = self.generate_children(samples)
    #
    #     subject = OrdinaryTest(random=random_string())
    #     subject.save()
    #
    #     m2m = M2MTest()
    #     m2m.save()
    #     m2m.relationship.add(*children)
    #     m2m.save()
    #     ModelEvent.objects.all().delete()
    #
    #     m2m.relationship.remove(subject)
    #     m2m.save()
    #
    #     events = ModelEvent.objects.all()
    #     # TODO: fails
    #     # self.assertEqual(events.count(), 0)

    def test_find(self):
        m2m = M2MTest()
        m2m.save()

        self.assertIsNotNone(find_m2m_rel(m2m.relationship.through, M2MTest))
        self.assertIsNone(find_m2m_rel(m2m.relationship.through, OrdinaryTest))

    def test_reverse(self):
        m2m = M2MTest()
        m2m.save()

        subject = OrdinaryTest(random=random_string())
        subject.save()

        ModelEvent.objects.all().delete()

        subject.m2mtest_set.add(m2m)
        subject.save()

        events = ModelEvent.objects.all()
        self.assertEqual(events.count(), 1)

        event = events[0]

        self.assertEqual(event.modifications.count(), 0)
        self.assertEqual(event.relationships.count(), 1)
        self.assertEqual(event.entry.mirror.name, 'M2MTest')

        relationship = event.relationships.all()[0]
        self.assertEqual(relationship.operation, int(Operation.CREATE))
        self.assertEqual(relationship.field.name, 'relationship')
        self.assertEqual(relationship.entry.primary_key, str(subject.id))


# TODO: test lazy_model_exclusion



from datetime import timedelta

from marshmallow import ValidationError

from automated_logging.signals import _function_model_exclusion
from automated_logging.tests.base import BaseTestCase


class MiscellaneousTestCase(BaseTestCase):
    def test_no_sender(self):
        self.assertIsNone(_function_model_exclusion(None, '', ''))

    def test_wrong_duration(self):
        from django.conf import settings
        from automated_logging.settings import settings as conf

        settings.AUTOMATED_LOGGING['unspecified']['max_age'] = complex(1, 1)
        conf.load.cache_clear()
        self.clear()

        self.assertRaises(ValidationError, conf.load)

        settings.AUTOMATED_LOGGING['unspecified']['max_age'] = (
            timedelta.max.total_seconds() + 1
        )
        conf.load.cache_clear()
        self.clear()

        self.assertRaises(ValidationError, conf.load)

        settings.AUTOMATED_LOGGING['unspecified']['max_age'] = 'Haha, error go brrr'
        conf.load.cache_clear()
        self.clear()

        self.assertRaises(ValidationError, conf.load)

    def test_unsupported_search_string(self):
        from django.conf import settings
        from automated_logging.settings import settings as conf

        settings.AUTOMATED_LOGGING['unspecified']['exclude']['applications'] = [
            'te:abc'
        ]
        conf.load.cache_clear()
        self.clear()

        self.assertRaises(ValidationError, conf.load)

        # settings.AUTOMATED_LOGGING['unspecified']['exclude']['applications'] = []
        conf.load.cache_clear()
        self.clear()

    def test_duration_none(self):
        from django.conf import settings
        from automated_logging.settings import settings as conf

        settings.AUTOMATED_LOGGING['unspecified']['max_age'] = None
        conf.load.cache_clear()
        self.clear()

        self.assertIsNone(conf.unspecified.max_age)



"""
Helper functions that are specifically used in the signals only.
"""

import re
from fnmatch import fnmatch
from functools import lru_cache
from pathlib import Path
from typing import List, Optional, Callable, Any

from automated_logging.helpers import (
    get_or_create_meta,
    get_or_create_thread,
    function2path,
    Operation,
)
from automated_logging.models import RequestEvent, UnspecifiedEvent
from automated_logging.settings import settings
from automated_logging.helpers.schemas import Search


# suboptimal meta is also cached -> look into how to solve
@lru_cache()
def cached_model_exclusion(sender, meta, operation) -> bool:
    """ cached so that we don't need to abuse ._meta and can invalidate the cache """
    return model_exclusion(sender, meta, operation)


def lazy_model_exclusion(instance, operation, sender) -> bool:
    """
    First look if the model has been excluded already
    -> only then look if excluded.

    Replaced by LRU-Cache.
    """

    return cached_model_exclusion(sender, instance._meta, operation)


def candidate_in_scope(candidate: str, scope: List[Search]) -> bool:
    """
    Check if the candidate string is valid with the scope supplied,
    the scope should be list of search strings - that can be either
    glob, plain or regex

    :param candidate: search string
    :param scope: List of Search
    :return: valid?
    """

    for search in scope:
        match = False
        if search.type == 'glob':
            match = fnmatch(candidate.lower(), search.value.lower())
        if search.type == 'plain':
            match = candidate.lower() == search.value.lower()
        if search.type == 'regex':
            match = bool(re.match(search.value, candidate, re.IGNORECASE))

        if match:
            return True

    return False


def request_exclusion(event: RequestEvent, view: Optional[Callable] = None) -> bool:
    """
    Determine if a request should be ignored/excluded from getting
    logged, these exclusions should be specified in the settings.

    :param event: RequestEvent
    :param view: Optional - function used by the resolver
    :return: should be excluded?
    """

    if view:
        thread, _ = get_or_create_thread()
        ignore = thread.dal['ignore.views']
        include = thread.dal['include.views']
        path = function2path(view)

        # if include None or method in include return False and don't
        # check further, else just continue with checking
        if path in include and (include[path] is None or event.method in include[path]):
            return False

        if (
            path in ignore
            # if ignored[compiled] is None, then no method will be ignored
            and ignore[path] is not None
            # ignored[compiled] == [] indicates all should be ignored
            and (len(ignore[path]) == 0 or event.method in ignore[path])
        ):
            return True

    exclusions = settings.request.exclude
    if event.method.lower() in exclusions.methods:
        return True

    if event.application.name and candidate_in_scope(
        event.application.name, exclusions.applications
    ):
        return True

    if event.status in exclusions.status:
        return True

    # if the application.name = None, then the application is unknown.
    # exclusions.unknown specifies if unknown should be excluded!
    if not event.application.name and exclusions.unknown:
        return True

    return False


def _function_model_exclusion(sender, scope: str, item: Any) -> Optional[bool]:
    if not sender:
        return None

    thread, _ = get_or_create_thread()
    ignore = thread.dal['ignore.models']
    include = thread.dal['include.models']

    path = function2path(sender)

    if path in include:
        items = getattr(include[path], scope)
        if items is None or item in items:
            return False

    if path in ignore:
        items = getattr(ignore[path], scope)
        if items is not None and (len(items) == 0 or item in items):
            return True

    return None


def model_exclusion(sender, meta, operation: Operation) -> bool:
    """
    Determine if the instance of a model should be excluded,
    these exclusions should be specified in the settings.

    :param meta:
    :param sender:
    :param operation:
    :return: should be excluded?
    """
    decorators = _function_model_exclusion(sender, 'operations', operation)
    if decorators is not None:
        return decorators

    if hasattr(sender, 'LoggingIgnore') and (
        getattr(sender.LoggingIgnore, 'complete', False)
        or {
            Operation.CREATE: 'create',
            Operation.MODIFY: 'modify',
            Operation.DELETE: 'delete',
        }[operation]
        in [o.lower() for o in getattr(sender.LoggingIgnore, 'operations', [])]
    ):
        return True

    exclusions = settings.model.exclude
    module = sender.__module__
    name = sender.__name__
    application = meta.app_label

    if (
        candidate_in_scope(name, exclusions.models)
        or candidate_in_scope(f'{module}.{name}', exclusions.models)
        or candidate_in_scope(f'{application}.{name}', exclusions.models)
    ):
        return True

    if candidate_in_scope(module, exclusions.models):
        return True

    if application and candidate_in_scope(application, exclusions.applications):
        return True

    # if there is no application string then we assume the model
    # location is unknown, if the flag exclude.unknown = True, then we just exclude
    if not application and exclusions.unknown:
        return True

    return False


def field_exclusion(field: str, instance, sender=None) -> bool:
    """
    Determine if the field of an instance should be excluded.
    """

    decorators = _function_model_exclusion(sender, 'fields', field)
    if decorators is not None:
        return decorators

    if hasattr(instance.__class__, 'LoggingIgnore') and (
        getattr(instance.__class__.LoggingIgnore, 'complete', False)
        or field in getattr(instance.__class__.LoggingIgnore, 'fields', [])
    ):
        return True

    exclusions = settings.model.exclude
    application = instance._meta.app_label
    model = instance.__class__.__name__

    if (
        candidate_in_scope(field, exclusions.fields)
        or candidate_in_scope(f'{model}.{field}', exclusions.fields)
        or candidate_in_scope(f'{application}.{model}.{field}', exclusions.fields)
    ):
        return True

    return False


def unspecified_exclusion(event: UnspecifiedEvent) -> bool:
    """
    Determine if an unspecified event needs to be excluded.
    """
    exclusions = settings.unspecified.exclude

    if event.application.name and candidate_in_scope(
        event.application.name, exclusions.applications
    ):
        return True

    if candidate_in_scope(str(event.file), exclusions.files):
        return True

    path = Path(event.file)
    # match greedily by first trying the complete path, if that doesn't match try
    # full relative and then complete relative.
    if [
        v
        for v in exclusions.files
        if v.type != 'regex'
        and (
            path.match(v.value)
            or path.match(f'/*{v.value}')
            or fnmatch(path, f'{v.value}/*')
            or fnmatch(path, f'/*{v.value}')
            or fnmatch(path, f'/*{v.value}/*')
        )
    ]:
        return True

    # application.name = None and exclusion.unknown = True
    if not event.application.name and exclusions.unknown:
        return True

    return False



import logging
import threading
from typing import NamedTuple, Optional, TYPE_CHECKING

from django.http import HttpRequest, HttpResponse

if TYPE_CHECKING:
    from django.contrib.auth.models import AbstractUser

RequestInformation = NamedTuple(
    'RequestInformation',
    [
        ('request', HttpRequest),
        ('response', Optional[HttpResponse]),
        ('exception', Optional[Exception]),
    ],
)


class AutomatedLoggingMiddleware:
    """
    Middleware used by django-automated-logging
    to provide request specific data to the request signals via
    the local thread.
    """

    thread = threading.local()

    def __init__(self, get_response):
        self.get_response = get_response

        AutomatedLoggingMiddleware.thread.__dal__ = None

    @staticmethod
    def save(request, response=None, exception=None):
        """
        Helper middleware, that sadly needs to be present.
        the request_finished and request_started signals only
        expose the class, not the actual request and response.

        We save the request and response specific data in the thread.

        :param request: Django Request
        :param response: Optional Django Response
        :param exception: Optional Exception
        :return:
        """

        AutomatedLoggingMiddleware.thread.__dal__ = RequestInformation(
            request, response, exception
        )

    def __call__(self, request):
        """
        TODO: fix staticfiles has no environment?!
        it seems like middleware isn't getting called for serving the static files,
        this seems very odd.

        There are 2 different states, request object will be stored when available
        and response will only be available post get_response.

        :param request:
        :return:
        """
        self.save(request)

        response = self.get_response(request)

        self.save(request, response)

        return response

    def process_exception(self, request, exception):
        """
        Exception proceeds the same as __call__ and therefore should
        also save things in the local thread.

        :param request: Django Request
        :param exception: Thrown Exception
        :return: -
        """
        self.save(request, exception=exception)

    @staticmethod
    def cleanup():
        """
        Cleanup function, that should be called last. Overwrites the
        custom __dal__ object with None, to make sure the next request
        does not use the same object.

        :return: -
        """
        AutomatedLoggingMiddleware.thread.__dal__ = None

    @staticmethod
    def get_current_environ() -> Optional[RequestInformation]:
        """
        Helper staticmethod that looks if the __dal__ custom attribute
        is present and returns either the attribute or None

        :return: Optional[RequestInformation]
        """

        if getattr(AutomatedLoggingMiddleware.thread, '__dal__', None):
            return RequestInformation(*AutomatedLoggingMiddleware.thread.__dal__)

        return None

    @staticmethod
    def get_current_user(
        environ: RequestInformation = None,
    ) -> Optional['AbstractUser']:
        """
        Helper staticmethod that returns the current user, taken from
        the current environment.

        :return: Optional[User]
        """
        from django.contrib.auth.models import AnonymousUser

        if not environ:
            environ = AutomatedLoggingMiddleware.get_current_environ()

        if not environ:
            return None

        if isinstance(environ.request.user, AnonymousUser):
            return None

        return environ.request.user



"""
Everything related to the admin interface of ModelEntry is located in here
"""

from django.contrib.admin import register
from django.shortcuts import redirect

from automated_logging.admin.model_event import ModelEventAdmin
from automated_logging.admin.base import ReadOnlyAdminMixin, ReadOnlyTabularInlineMixin
from automated_logging.models import ModelEntry, ModelEvent


class ModelEventInline(ReadOnlyTabularInlineMixin):
    """ inline for all attached events """

    model = ModelEvent

    def __init__(self, *args, **kwargs):
        super(ModelEventInline, self).__init__(*args, **kwargs)

        self.readonly_fields = [*self.readonly_fields, 'get_uuid', 'get_modifications']

    def get_uuid(self, instance):
        """ make the uuid small """
        return self.model_admin_url(instance, str(instance.id).split('-')[0])

    get_uuid.short_description = 'UUID'

    def get_modifications(self, instance):
        """ ModelEventAdmin already implements this functions, we just refer to it"""
        return ModelEventAdmin.get_modifications(self, instance)

    get_modifications.short_description = 'Modifications'

    fields = ('get_uuid', 'updated_at', 'user', 'get_modifications')

    ordering = ('-updated_at',)

    verbose_name = 'Event'
    verbose_name_plural = 'Events'


@register(ModelEntry)
class ModelEntryAdmin(ReadOnlyAdminMixin):
    """ admin page specification for ModelEntry """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.readonly_fields = [*self.readonly_fields, 'get_model', 'get_application']

    def changelist_view(self, request, **kwargs):
        """ instead of showing the changelist view redirect to the parent app_list"""
        return redirect('admin:app_list', self.model._meta.app_label)

    def has_module_permission(self, request):
        """ remove model entries from the index.html list """
        return False

    def get_model(self, instance):
        """ get the model mirror """
        return self.model_admin_url(instance.mirror)

    get_model.short_description = 'Model'

    def get_application(self, instance):
        """ get the application """
        return instance.mirror.application

    get_application.short_description = 'Application'

    fieldsets = (
        (
            'Information',
            {'fields': ('id', 'get_model', 'get_application', 'primary_key', 'value')},
        ),
    )

    inlines = [ModelEventInline]



import uuid
from django.db.models import (
    Model,
    UUIDField,
    DateTimeField,
    ManyToManyField,
    CASCADE,
    ForeignKey,
    OneToOneField,
    CharField,
)

from automated_logging.decorators import exclude_model, include_model


class TestBase(Model):
    """ Base for all the test models """

    id = UUIDField(default=uuid.uuid4, primary_key=True)

    created_at = DateTimeField(auto_now_add=True)
    updated_at = DateTimeField(auto_now=True)

    class Meta:
        abstract = True
        app_label = 'automated_logging'


class OrdinaryBaseTest(TestBase):
    """ Ordinary base test. Has a random char field."""

    random = CharField(max_length=255, null=True)
    random2 = CharField(max_length=255, null=True)

    class Meta:
        abstract = True
        app_label = 'automated_logging'


class OrdinaryTest(OrdinaryBaseTest):
    """ Ordinary test. Has a random char field."""

    class Meta:
        app_label = 'automated_logging'


class M2MTest(TestBase):
    """ Used to test the Many-To-Many Relationship functionality of DAL"""

    relationship = ManyToManyField(OrdinaryTest)

    class Meta:
        app_label = 'automated_logging'


class ForeignKeyTest(TestBase):
    """ Used to test ForeignKey functionality of DAL."""

    relationship = ForeignKey(OrdinaryTest, on_delete=CASCADE, null=True)

    class Meta:
        app_label = 'automated_logging'


class OneToOneTest(TestBase):
    """ Used to test the One-To-One Relationship functionality of DAL."""

    relationship = OneToOneField(OrdinaryTest, on_delete=CASCADE, null=True)

    class Meta:
        app_label = 'automated_logging'


class SpeedTest(TestBase):
    """ Used to test the speed of DAL """

    for idx in range(100):
        exec(f"column{idx} = CharField(max_length=15, null=True)")

    class Meta:
        app_label = 'automated_logging'


class FullClassBasedExclusionTest(OrdinaryBaseTest):
    """ Used to test the full model exclusion via meta class"""

    class Meta:
        app_label = 'automated_logging'

    class LoggingIgnore:
        complete = True


class PartialClassBasedExclusionTest(OrdinaryBaseTest):
    """ Used to test partial ignore via fields """

    class Meta:
        app_label = 'automated_logging'

    class LoggingIgnore:
        fields = ['random']
        operations = ['delete']


@exclude_model
class FullDecoratorBasedExclusionTest(OrdinaryBaseTest):
    """ Used to test full decorator exclusion """

    class Meta:
        app_label = 'automated_logging'


@exclude_model(operations=['delete'], fields=['random'])
class PartialDecoratorBasedExclusionTest(OrdinaryBaseTest):
    """ Used to test partial decorator exclusion """

    class Meta:
        app_label = 'automated_logging'


@include_model
class DecoratorOverrideExclusionTest(OrdinaryBaseTest):
    """
    Used to check if include_model
    has precedence over class based configuration
    """

    class Meta:
        app_label = 'automated_logging'

    class LoggingIgnore:
        complete = True



"""
File handles every signal related to the saving/deletion of django models.
"""

import logging
from collections import namedtuple
from datetime import datetime
from pprint import pprint
from typing import Any

from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction
from django.db.models.signals import pre_save, post_save, post_delete
from django.dispatch import receiver

from automated_logging.models import (
    ModelValueModification,
    ModelField,
    ModelMirror,
    Application,
)
from automated_logging.settings import settings
from automated_logging.signals import (
    model_exclusion,
    lazy_model_exclusion,
    field_exclusion,
)
from automated_logging.helpers import (
    get_or_create_meta,
    Operation,
    get_or_create_model_event,
)
from automated_logging.helpers.enums import PastOperationMap

ChangeSet = namedtuple('ChangeSet', ('deleted', 'added', 'changed'))
logger = logging.getLogger(__name__)


def normalize_save_value(value: Any):
    """ normalize the values given to the function to make stuff more readable """
    if value is None or value == '':
        return None
    if isinstance(value, str):
        return value

    return repr(value)


@receiver(pre_save, weak=False)
@transaction.atomic
def pre_save_signal(sender, instance, **kwargs) -> None:
    """
    Compares the current instance and old instance (fetched via the pk)
    and generates a dictionary of changes

    :param sender:
    :param instance:
    :param kwargs:
    :return: None
    """
    get_or_create_meta(instance)
    # clear the event to be sure
    instance._meta.dal.event = None

    operation = Operation.MODIFY
    try:
        pre = sender.objects.get(pk=instance.pk)
    except ObjectDoesNotExist:
        # __dict__ is used on pre, therefore we need to create a function
        # that uses __dict__ too, but returns nothing.

        pre = lambda _: None
        operation = Operation.CREATE

    excluded = lazy_model_exclusion(instance, operation, instance.__class__)
    if excluded:
        return

    old, new = pre.__dict__, instance.__dict__

    previously = set(
        k for k in old.keys() if not k.startswith('_') and old[k] is not None
    )
    currently = set(
        k for k in new.keys() if not k.startswith('_') and new[k] is not None
    )

    added = currently.difference(previously)
    deleted = previously.difference(currently)
    changed = (
        set(
            k
            for k, v in set(
                # generate a set from the dict of old values
                # exclude protected and magic attributes
                (k, v)
                for k, v in old.items()
                if not k.startswith('_')
            ).difference(
                set(
                    # generate a set from the dict of new values
                    # also exclude the protected and magic attributes
                    (k, v)
                    for k, v in new.items()
                    if not k.startswith('_')
                )
            )
            # the difference between the two sets
            # because the keys are the same, but values are different
            # will result in a change set of changed values
            # ('a', 'b') in old and new will get eliminated
            # ('a', 'b') in old and ('a', 'c') in new will
            # result in ('a', 'b') in the new set as that is
            # different.
        )
        # remove all added and deleted attributes from the changelist
        # they would still be present, because None -> Value
        .difference(added).difference(deleted)
    )

    summary = [
        *(
            {
                'operation': Operation.CREATE,
                'previous': None,
                'current': new[k],
                'key': k,
            }
            for k in added
        ),
        *(
            {
                'operation': Operation.DELETE,
                'previous': old[k],
                'current': None,
                'key': k,
            }
            for k in deleted
        ),
        *(
            {
                'operation': Operation.MODIFY,
                'previous': old[k],
                'current': new[k],
                'key': k,
            }
            for k in changed
        ),
    ]

    # exclude fields not present in _meta.get_fields
    fields = {f.name: f for f in instance._meta.get_fields()}
    extra = {f.attname: f for f in instance._meta.get_fields() if hasattr(f, 'attname')}
    fields = {**extra, **fields}

    summary = [s for s in summary if s['key'] in fields.keys()]

    # field exclusion
    summary = [
        s
        for s in summary
        if not field_exclusion(s['key'], instance, instance.__class__)
    ]

    model = ModelMirror()
    model.name = sender.__name__
    model.application = Application(name=instance._meta.app_label)

    modifications = []
    for entry in summary:
        field = ModelField()
        field.name = entry['key']
        field.mirror = model

        field.type = fields[entry['key']].__class__.__name__

        modification = ModelValueModification()
        modification.operation = entry['operation']
        modification.field = field

        modification.previous = normalize_save_value(entry['previous'])
        modification.current = normalize_save_value(entry['current'])

        modifications.append(modification)

    instance._meta.dal.modifications = modifications

    if settings.model.performance:
        instance._meta.dal.performance = datetime.now()


def post_processor(status, sender, instance, updated=None, suffix='') -> None:
    """
    Due to the fact that both post_delete and post_save have
    the same logic for propagating changes, we have this helper class
    to do so, just simply wraps and logs the data the handler needs.

    The event gets created here instead of the handler to keep
    everything consistent and have the handler as simple as possible.

    :param status: Operation
    :param sender: model class
    :param instance: model instance
    :param updated: updated fields
    :param suffix: suffix to be added to the message
    :return: None
    """
    past = {v: k for k, v in PastOperationMap.items()}

    get_or_create_meta(instance)

    event, _ = get_or_create_model_event(instance, status, force=True, extra=True)
    modifications = getattr(instance._meta.dal, 'modifications', [])

    # clear the modifications meta list
    instance._meta.dal.modifications = []

    if len(modifications) == 0 and status == Operation.MODIFY:
        # if the event is modify, but nothing changed, don't actually propagate
        return

    logger.log(
        settings.model.loglevel,
        f'{event.user or "Anonymous"} {past[status]} '
        f'{event.entry.mirror.application}.{sender.__name__} | '
        f'Instance: {instance!r}{suffix}',
        extra={
            'action': 'model',
            'data': {'status': status, 'instance': instance},
            'event': event,
            'modifications': modifications,
        },
    )


@receiver(post_save, weak=False)
@transaction.atomic
def post_save_signal(
    sender, instance, created, update_fields: frozenset, **kwargs
) -> None:
    """
    Signal is getting called after a save has been concluded. When this
    is the case we can be sure the save was successful and then only
    propagate the changes to the handler.

    :param sender: model class
    :param instance: model instance
    :param created: bool, was the model created?
    :param update_fields: which fields got explicitly updated?
    :param kwargs: django needs kwargs to be there
    :return: -
    """
    status = Operation.CREATE if created else Operation.MODIFY
    if lazy_model_exclusion(instance, status, instance.__class__,):
        return
    get_or_create_meta(instance)

    suffix = f''
    if (
        status == Operation.MODIFY
        and hasattr(instance._meta.dal, 'modifications')
        and settings.model.detailed_message
    ):
        suffix = (
            f' | Modifications: '
            f'{", ".join([m.short() for m in instance._meta.dal.modifications])}'
        )

    if update_fields is not None and hasattr(instance._meta.dal, 'modifications'):
        instance._meta.dal.modifications = [
            m for m in instance._meta.dal.modifications if m.field.name in update_fields
        ]

    post_processor(status, sender, instance, update_fields, suffix)


@receiver(post_delete, weak=False)
@transaction.atomic
def post_delete_signal(sender, instance, **kwargs) -> None:
    """
    Signal is getting called after instance deletion. We just redirect the
    event to the post_processor.

    TODO: consider doing a "delete snapshot"

    :param sender: model class
    :param instance: model instance
    :param kwargs: required bt django
    :return: -
    """

    get_or_create_meta(instance)
    instance._meta.dal.event = None

    if lazy_model_exclusion(instance, Operation.DELETE, instance.__class__):
        return

    post_processor(Operation.DELETE, sender, instance)



""" custom exceptions """


class NoMatchFound(Exception):
    """ error that indicates that no match has been found for a regex """

    pass


class CouldNotConvertError(Exception):
    """ error that is thrown when no conversion could be done """

    pass



# Generated by Django 3.0.7 on 2020-08-19 10:04

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('automated_logging', '0016_auto_20200803_1917'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='application',
            options={'verbose_name': 'Application', 'verbose_name_plural': 'Applications'},
        ),
    ]



""" test specific helpers """

import string
from random import choice


def random_string(length=10):
    """ generate a random string with the length specified """
    return ''.join(choice(string.ascii_letters) for _ in range(length))



"""
Everything related to the admin interface of RequestEvent is located in here
"""
from django.contrib.admin import register, RelatedOnlyFieldListFilter

from automated_logging.admin.base import ReadOnlyAdminMixin, ReadOnlyTabularInlineMixin
from automated_logging.models import RequestEvent


@register(RequestEvent)
class RequestEventAdmin(ReadOnlyAdminMixin):
    """ admin page specification for the RequestEvent """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.readonly_fields = [*self.readonly_fields, 'get_uri', 'get_user']

    def get_id(self, instance):
        """ shorten the id to the first 8 digits """
        return str(instance.id).split('-')[0]

    get_id.short_description = 'UUID'

    def get_uri(self, instance):
        """ get the uri. just a redirect to set the short description. """
        return instance.uri

    get_uri.short_description = 'URI'

    def get_user(self, instance):
        """ get the user with a URL """
        return self.model_admin_url(instance.user)

    get_user.short_description = 'User'

    list_display = ('get_id', 'updated_at', 'user', 'method', 'status', 'uri')

    date_hierarchy = 'updated_at'
    ordering = ('-updated_at',)

    list_filter = ('updated_at', ('user', RelatedOnlyFieldListFilter))

    fieldsets = (
        ('Information', {'fields': ('id', 'get_user', 'updated_at', 'application',)},),
        ('HTTP', {'fields': ('method', 'status', 'get_uri')}),
    )
    # TODO: Context



from django.apps import AppConfig
from .settings import settings


class AutomatedloggingConfig(AppConfig):
    name = 'automated_logging'
    verbose_name = 'Django Automated Logging (DAL)'

    def ready(self):
        if 'request' in settings.modules:
            from .signals import request
        if 'model' in settings.modules:
            from .signals import save
            from .signals import m2m

        from .handlers import DatabaseHandler



""" Various enums used in DAL """

from enum import Enum


class Operation(int, Enum):
    """
    Simple Enum that will be used across the code to
    indicate the current operation that happened.

    Due to the fact that enum support for django was
    only added in 3.0 we have DjangoOperations to convert
    it to the old django format.
    """

    CREATE = 1
    MODIFY = 0
    DELETE = -1


DjangoOperations = [(e.value, o.lower()) for o, e in Operation.__members__.items()]
VerbOperationMap = {
    'create': Operation.CREATE,
    'modify': Operation.MODIFY,
    'delete': Operation.DELETE,
    'add': Operation.CREATE,
    'remove': Operation.DELETE,
}
VerbM2MOperationMap = {
    'add': Operation.CREATE,
    'modify': Operation.MODIFY,
    'remove': Operation.DELETE,
}
PastOperationMap = {
    'created': Operation.CREATE,
    'modified': Operation.MODIFY,
    'deleted': Operation.DELETE,
}
PastM2MOperationMap = {
    'added': Operation.CREATE,
    'modified': Operation.MODIFY,
    'removed': Operation.DELETE,
}
ShortOperationMap = {
    '+': Operation.CREATE,
    '~': Operation.MODIFY,
    '-': Operation.DELETE,
}
TranslationOperationMap = {**VerbOperationMap, **PastOperationMap, **ShortOperationMap}



"""
Everything related to the admin interface of ModelEvent is located in here
"""

from django.contrib.admin import register, RelatedOnlyFieldListFilter
from django.utils.html import format_html

from automated_logging.admin.base import ReadOnlyTabularInlineMixin, ReadOnlyAdminMixin
from automated_logging.helpers import Operation
from automated_logging.models import (
    ModelValueModification,
    ModelRelationshipModification,
    ModelEvent,
)


class ModelValueModificationInline(ReadOnlyTabularInlineMixin):
    """ inline for all modifications """

    model = ModelValueModification

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.readonly_fields = [*self.readonly_fields, 'get_uuid', 'get_field']

    def get_uuid(self, instance):
        """ make the uuid small """
        return str(instance.id).split('-')[0]

    get_uuid.short_description = 'UUID'

    def get_field(self, instance):
        """ show the field name """
        return instance.field.name

    get_field.short_description = 'Field'

    fields = ('get_uuid', 'operation', 'get_field', 'previous', 'current')
    can_delete = False

    verbose_name = 'Modification'
    verbose_name_plural = 'Modifications'


class ModelRelationshipModificationInline(ReadOnlyTabularInlineMixin):
    """ inline for all relationship modifications """

    model = ModelRelationshipModification

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.readonly_fields = [*self.readonly_fields, 'get_uuid', 'get_field']

    def get_uuid(self, instance):
        """ make the uuid small """
        return str(instance.id).split('-')[0]

    get_uuid.short_description = 'UUID'

    def get_field(self, instance):
        """ show the field name """
        return instance.field.name

    get_field.short_description = 'Field'

    fields = ('get_uuid', 'operation', 'get_field', 'model')
    can_delete = False

    verbose_name = 'Relationship'
    verbose_name_plural = 'Relationships'


@register(ModelEvent)
class ModelEventAdmin(ReadOnlyAdminMixin):
    """ admin page specification for ModelEvent """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.readonly_fields = [
            *self.readonly_fields,
            'get_application',
            'get_user',
            'get_model_link',
        ]

    def get_modifications(self, instance):
        """
        Modifications in short form, are colored for better readability.

        Colors taken from
        https://github.com/django/django/tree/master/django/contrib/admin/static/admin/img
        """
        colors = {
            Operation.CREATE: '#70bf2b',
            Operation.MODIFY: '#efb80b',
            Operation.DELETE: '#dd4646',
        }
        return format_html(
            ', '.join(
                [
                    *[
                        f'<span style="color: {colors[Operation(m.operation)]};">'
                        f'{m.short()}'
                        f'</span>'
                        for m in instance.modifications.all()
                    ],
                    *[
                        f'<span style="color: {colors[Operation(r.operation)]};">'
                        f'{r.medium()[0]}'
                        f'</span>[{r.medium()[1]}]'
                        for r in instance.relationships.all()
                    ],
                ],
            )
        )

    get_modifications.short_description = 'Modifications'

    def get_model(self, instance):
        """
        get the model
        TODO: consider splitting this up to model/pk/value
        """
        return instance.entry.short()

    get_model.short_description = 'Model'

    def get_model_link(self, instance):
        """ get the model with a link to the entry """
        return self.model_admin_url(instance.entry)

    get_model_link.short_description = 'Model'

    def get_application(self, instance):
        """
        helper to get the application from the child ModelMirror
        :param instance:
        :return:
        """
        return instance.entry.mirror.application

    get_application.short_description = 'Application'

    def get_id(self, instance):
        """ shorten the id to the first 8 digits """
        return str(instance.id).split('-')[0]

    get_id.short_description = 'UUID'

    def get_user(self, instance):
        """ return the user with a link """
        return self.model_admin_url(instance.user) if instance.user else None

    get_user.short_description = 'User'

    list_display = (
        'get_id',
        'updated_at',
        'user',
        'get_application',
        'get_model',
        'get_modifications',
    )

    list_filter = (
        'updated_at',
        ('user', RelatedOnlyFieldListFilter),
        ('model__model__application', RelatedOnlyFieldListFilter),
        ('model__model', RelatedOnlyFieldListFilter),
    )

    date_hierarchy = 'updated_at'
    ordering = ('-updated_at',)

    fieldsets = (
        (
            'Information',
            {
                'fields': (
                    'id',
                    'get_user',
                    'updated_at',
                    'get_application',
                    'get_model_link',
                )
            },
        ),
        ('Introspection', {'fields': ('performance', 'snapshot')}),
    )
    inlines = [ModelValueModificationInline, ModelRelationshipModificationInline]

    show_change_link = True



""" Test base every unit test uses """
import importlib
import logging.config
from copy import copy, deepcopy

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AbstractUser
from django.test import TestCase, RequestFactory
from django.urls import path

from automated_logging.helpers import namedtuple2dict
from automated_logging.middleware import AutomatedLoggingMiddleware
from automated_logging.models import ModelEvent, RequestEvent, UnspecifiedEvent
from automated_logging.signals import cached_model_exclusion

User: AbstractUser = get_user_model()
USER_CREDENTIALS = {'username': 'example', 'password': 'example'}


class BaseTestCase(TestCase):
    def __init__(self, method_name):
        from django.conf import settings

        settings.AUTOMATED_LOGGING_DEV = True

        super().__init__(method_name)

    def request(self, method, view, data=None):
        """
        request a specific view and return the response.

        This is not ideal and super hacky. Backups the actual urlpatterns,
        and then overrides the urlpatterns with a temporary one and then
        inserts the new one again.
        """

        urlconf = importlib.import_module(settings.ROOT_URLCONF)

        backup = copy(urlconf.urlpatterns)
        urlconf.urlpatterns.clear()
        urlconf.urlpatterns.append(path('', view))

        response = self.client.generic(method, '/', data=data)

        urlconf.urlpatterns.clear()
        urlconf.urlpatterns.extend(backup)

        return response

    def setUp(self):
        """ setUp the DAL specific test environment """
        from django.conf import settings
        from automated_logging.settings import default, settings as conf

        self.user = User.objects.create_user(**USER_CREDENTIALS)
        self.user.save()

        self.original_config = deepcopy(settings.AUTOMATED_LOGGING)

        base = namedtuple2dict(default)

        settings.AUTOMATED_LOGGING.clear()
        for key, value in base.items():
            settings.AUTOMATED_LOGGING[key] = deepcopy(value)

        conf.load.cache_clear()

        self.setUpLogging()
        super().setUp()

    def tearDown(self) -> None:
        """ tearDown the DAL specific environment """
        from django.conf import settings
        from automated_logging.settings import settings as conf

        super().tearDown()

        self.tearDownLogging()

        settings.AUTOMATED_LOGGING.clear()
        for key, value in self.original_config.items():
            settings.AUTOMATED_LOGGING[key] = deepcopy(value)

        conf.load.cache_clear()

        if hasattr(AutomatedLoggingMiddleware.thread, 'dal'):
            delattr(AutomatedLoggingMiddleware.thread, 'dal')

        cached_model_exclusion.cache_clear()

    @staticmethod
    def clear():
        """ clear all events """
        ModelEvent.objects.all().delete()
        RequestEvent.objects.all().delete()
        UnspecifiedEvent.objects.all().delete()

    def tearDownLogging(self):
        """
        replace our own logging config
        with the original to not break any other tests
        that might depend on it.
        """
        from django.conf import settings

        settings.LOGGING = self.logging_backup
        logging.config.dictConfig(settings.LOGGING)

    def setUpLogging(self):
        """ sets up logging dict, so that we can actually use our own """
        from django.conf import settings

        self.logging_backup = deepcopy(settings.LOGGING)
        settings.LOGGING = {
            'version': 1,
            'disable_existing_loggers': False,
            'root': {'level': 'INFO', 'handlers': ['console', 'db'],},
            'formatters': {
                'verbose': {
                    'format': '%(levelname)s %(asctime)s %(module)s '
                    '%(process)d %(thread)d %(message)s'
                },
                'simple': {'format': '%(levelname)s %(message)s'},
                'syslog': {
                    'format': '%(asctime)s %%LOCAL0-%(levelname) %(message)s'
                    # 'format': '%(levelname)s %(message)s'
                },
            },
            'handlers': {
                'console': {
                    'level': 'INFO',
                    'class': 'logging.StreamHandler',
                    'formatter': 'verbose',
                },
                'db': {
                    'level': 'INFO',
                    'class': 'automated_logging.handlers.DatabaseHandler',
                },
            },
            'loggers': {
                'automated_logging': {
                    'level': 'INFO',
                    'handlers': ['console', 'db'],
                    'propagate': False,
                },
                'django': {
                    'level': 'INFO',
                    'handlers': ['console', 'db'],
                    'propagate': False,
                },
            },
        }

        logging.config.dictConfig(settings.LOGGING)

    def bypass_request_restrictions(self):
        """ bypass all request default restrictions of DAL """
        from django.conf import settings
        from automated_logging.settings import settings as conf

        settings.AUTOMATED_LOGGING['request']['exclude']['status'] = []
        settings.AUTOMATED_LOGGING['request']['exclude']['methods'] = []
        conf.load.cache_clear()

        self.clear()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['automated_logging',
 'automated_logging.admin',
 'automated_logging.helpers',
 'automated_logging.migrations',
 'automated_logging.signals',
 'automated_logging.tests']

package_data = \
{'': ['*'], 'automated_logging': ['templates/dal/admin/*']}

install_requires = \
['django-ipware>=3.0.0,<4.0.0',
 'django-picklefield>=3.0.1,<4.0.0',
 'django>=2.2,<4.0',
 'marshmallow>=3.6.1,<4.0.0']

setup_kwargs = {
    'name': 'django-automated-logging',
    'version': '6.0.2',
    'description': 'Django Database Based Automated Logging - finally solved and done in a proper way.',
    'long_description': '# Django Database Based Automated Logging\n\n![https://pypi.python.org/pypi?name=django-automated-logging](https://img.shields.io/pypi/v/django-automated-logging.svg)\n![https://pypi.python.org/pypi?name=django-automated-logging](https://img.shields.io/pypi/l/django-automated-logging.svg)\n![https://pypi.python.org/pypi?name=django-automated-logging](https://img.shields.io/pypi/status/django-automated-logging.svg)\n![https://pypi.python.org/pypi?name=django-automated-logging](https://img.shields.io/pypi/pyversions/django-automated-logging.svg)\n![https://travis-ci.org/indietyp/django-automated-logging](https://travis-ci.org/indietyp/django-automated-logging.svg?branch=master)\n![https://coveralls.io/github/indietyp/django-automated-logging?branch=master](https://coveralls.io/repos/github/indietyp/django-automated-logging/badge.svg?branch=master)\n![https://www.codacy.com/app/bilalmahmoud/django-automated-logging?utm_source=github.com&amp;utm_medium=referral&amp;utm_content=indietyp/django-automated-logging&amp;utm_campaign=Badge_Grade](https://api.codacy.com/project/badge/Grade/96fdb764fc34486399802b2f8267efcc)\n![https://paypal.me/indietyp/5](https://img.shields.io/badge/Support%20the%20Project-PayPal-green.svg)\n\n**Notice:** Most of this will be moved into a wiki.\n\n## What is Django-Automated-Logging (DAL)?\n\nTL;DR: DAL is a package to **automatically** track changes in your project, ranging\nfrom simple logging messages, to model changes or requests done by users.\n\nYou can decide what you want to do and how.\nDAL allows fine-grained customization and filtering with various methods.\n\n### Introduction\n\nDjango Fully Automated Logging - **finally** solved and done properly.\n\nHow to install?\n`pip install django-automated-logging` or `poetry add django-automated-logging`\n\n### What is the purpose?\nThe goal of DAL is to provide an easy, accessible and DRY way to log the inner working of you applications.\nUltimately giving you the chance to easily see what is happening without excessive manual print/logging statements.\n\nThe application uses minimal requirements and is performant.\n\n### How does it work?\nThe application facilitates the built-in logging mechanic\nby providing a custom handler, that just needs to be added to the `LOGGING` configuration.\n\nDAL uses native Django signals to know what is happening behind the scenes without injecting custom code.\n\n### Minimal Setup\n\nYou can also configure DAL to only log to a file and not to a database.\nYou just need to enable DAL and not include the custom logging handler.\n\n## Detailed Information\n\n### Features\n\n1. Easy Setup\n2. Extensible\n3. Feature-Rich\n4. Completely Automated\n5. Built-In Database Logger\n6. No custom code needs to be inserted into your codebase\n7. Can capture logging messages unrelated to the package itself\n8. Only does what it needs to do, no extra bells and whistles.\n\n### Setup\n\nInitial Configuration is via your projects `settings.py`\n\n1. `INSTALLED_APPS` append: `\'automated_logging\'`\n2. `MIDDLEWARE` append: `\'automated_logging.middleware.AutomatedLoggingMiddleware\'`\n3. `LOGGING` section `handlers` add:\n    ```python\n       \'db\': {\n           \'level\': \'INFO\',\n           \'class\': \'automated_logging.handlers.DatabaseHandler\',\n       {\n    ```\n4. `LOGGING` section `loggers` add: (only required if database logging desired)\n    ```python\n       \'automated_logging\': {\n           \'level\': \'INFO\',\n           \'handlers\': [\'db\'],\n           \'propagate\': True,\n       },\n       \'django\': {\n           \'level\': \'INFO\',\n           \'handlers\': [\'db\'],\n           \'propagate\': True,\n       },\n    ```\n5. execute: `python manage.py migrate automated_logging`\n\n`LOGGING` configuration details are just recommendations.\n\n### Configuration\n\nFurther configuration can be done via the variable `AUTOMATED_LOGGING`. The defaults are:\n\n```python\nAUTOMATED_LOGGING = {\n    "globals": {\n        "exclude": {\n            "applications": [\n                "plain:contenttypes",\n                "plain:admin",\n                "plain:basehttp",\n                "glob:session*",\n                "plain:migrations",\n            ]\n        }\n    },\n    "model": {\n        "detailed_message": True,\n        "exclude": {"applications": [], "fields": [], "models": [], "unknown": False},\n        "loglevel": 20,\n        "mask": [],\n        "max_age": None,\n        "performance": False,\n        "snapshot": False,\n        "user_mirror": False,\n    },\n    "modules": ["request", "unspecified", "model"],\n    "request": {\n        "data": {\n            "content_types": ["application/json"],\n            "enabled": [],\n            "ignore": [],\n            "mask": ["password"],\n            "query": False,\n        },\n        "exclude": {\n            "applications": [],\n            "methods": ["GET"],\n            "status": [200],\n            "unknown": False,\n        },\n        "ip": True,\n        "loglevel": 20,\n        "max_age": None,\n    },\n    "unspecified": {\n        "exclude": {"applications": [], "files": [], "unknown": False},\n        "loglevel": 20,\n        "max_age": None,\n    },\n}\n```\n\nYou can always inspect the current default configuration by doing:\n\n```python\nfrom pprint import pprint\nfrom automated_logging.settings import default\nfrom automated_logging.helpers import namedtuple2dict\n\npprint(namedtuple2dict(default))\n```\n\n**Recommendation:** include the `globals` application defaults as those modules can be particularly verbose or be duplicates.\n\nThere are *three* different independent modules available `request` (for request logging), `unspecified` (for general logging messages), and `models` (for model changes).\nThey can be enabled and disabled by including them in the `modules` configuration.\n\nThe `loglevel` setting indicates the severity for the logging messages sent from the module.\n`INFO (20)` or `DEBUG (10)` is the right call for most cases.\n\n*New in 6.x.x:* Saving can be batched via the `batch` setting for the handler.\n\n*New in 6.x.x:* Saving can be threaded by `thread: True` for the handler settings. **This is highly experimental**\n\n*New in 6.x.x:* every field in `exclude` can be either be a `glob` (prefixing the string with `gl:`), a `regex` (prefixing the string with `re:`) or plain (prefixing the string with `pl:`). The default is `glob`.\n\n### Decorators\n\nYou can explicitly exclude or include views/models, by using the new decorators.\n\n```python\nfrom automated_logging.decorators import include_view, include_model, exclude_view, exclude_model\n\n@include_view(methods=None)\n@exclude_view(methods=[])\ndef view(request):\n    pass\n\n@include_model(operations=None, fields=None)\n@exclude_model(operations=[], fields=[])\nclass ExampleModel:\n    pass\n```\n\n`include` *always* takes precedence over `exclude`, if you use multiple `include` or `exclude` instead of overwriting they will *update/extend* the previous definition.\n\n`operations` can be either `create`, `modify`, `delete`. `fields` is a list model specific fields to be included/excluded.\n`methods` is a list methods to be included/excluded.\n\n### Class-Based Configuration\n\nClass-Based Configuration is done over a specific meta class `LoggingIgnore`. Decorators take precedence over class-based configuration, but class-based configuration takes precedence over AUTOMATED_LOGGING configuration.\n\n```python\nclass ExampleModel:\n    class LoggingIgnore:\n        complete = False\n        fields = []\n        operations = []\n```\n\nas described above `operations` and `fields` work the same way. `complete = True` means that a model is excluded no matter what.\n\n## Changelog\n\n### Version 6.0.0\n- **Added:** ``batch`` settings to the handler\n- **Added:** decorators\n- **Added:** class-based configuration\n- **Added:** request and response bodies can now be saved\n- **Added:** regex, glob matching for settings\n- **Updated:** settings\n- **Updated:** models\n- **Updated:** to current django version (2.2, 3.0, 3.1)\n- **Updated:** DAL no longer stores internal information directly, but now has a custom _meta object injected.\n- **Updated:** project now uses black for formatting\n- **Updated:** internals were completely rewritten for greater maintainability and speed.\n- **Fixed:** https://github.com/indietyp/django-automated-logging/issues/1\n- **Fixed:** https://github.com/indietyp/django-automated-logging/issues/2\n- **Moved:** `max_age` is now part of the `settings.py` configuration.\n\n### Version 5.0.0\n- **Added:** ``maxage`` handler setting to automatically remove database entries after a certain amount of time.\n- **Added:** query string in requests can now be enabled/disabled (are now disabled by default)\n- **Fixed:** Value and URI could be longer than 255 characters. DAL would throw an exception. This is fixed.\n\n\n## Roadmap\n\n### Version 6.1.x\n- [ ] archive options\n- [ ] decorators greater flexibility\n- [ ] wiki -> documentation\n- [ ] make django-ipware optional via extras\n- [ ] and more!\n\n### Version 7.x.x\n- [ ] implementation of a git like versioning interface\n\n### Version 8.x.x\n- [ ] temporary world domination',
    'author': 'Bilal Mahmoud',
    'author_email': 'bilalmahmoud@posteo.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/indietyp/django-automated-logging',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4',
}


setup(**setup_kwargs)



import re
from datetime import timedelta
from logging import Handler, LogRecord
from pathlib import Path
from threading import Thread
from typing import Dict, Any, TYPE_CHECKING, List, Optional, Union, Type, Tuple

from django.core.exceptions import ObjectDoesNotExist
from django.utils import timezone
from django.db.models import ForeignObject, Model


if TYPE_CHECKING:
    # we need to do this, to avoid circular imports
    from automated_logging.models import (
        RequestEvent,
        ModelEvent,
        ModelValueModification,
        ModelRelationshipModification,
    )


class DatabaseHandler(Handler):
    def __init__(
        self, *args, batch: Optional[int] = 1, threading: bool = False, **kwargs
    ):
        self.limit = batch or 1
        self.threading = threading
        self.instances = set()
        super(DatabaseHandler, self).__init__(*args, **kwargs)

    def save(self, instance=None):
        """
        Internal save procedure.
        Handles deletion when an event exceeds max_age
        and batch saving via atomic transactions.

        :return: None
        """
        from django.db import transaction
        from automated_logging.models import ModelEvent, RequestEvent, UnspecifiedEvent
        from automated_logging.settings import settings

        if instance:
            self.instances.add(instance)
        if len(self.instances) < self.limit:
            return

        def database(instances, config):
            """ wrapper so that we can actually use threading """
            with transaction.atomic():
                [i.save() for i in instances]

                current = timezone.now()
                if config.model.max_age:
                    ModelEvent.objects.filter(
                        created_at__lte=current - config.model.max_age
                    ).delete()
                if config.unspecified.max_age:
                    UnspecifiedEvent.objects.filter(
                        created_at__lte=current - config.unspecified.max_age
                    ).delete()

                if config.request.max_age:
                    RequestEvent.objects.filter(
                        created_at__lte=current - config.request.max_age
                    ).delete()
                instances.clear()

        if self.threading:
            thread = Thread(
                group=None, target=database, args=(self.instances, settings)
            )
            thread.start()
        else:
            database(self.instances, settings)

    def get_or_create(self, target: Type[Model], **kwargs) -> Tuple[Model, bool]:
        """
        proxy for "get_or_create" from django,
        instead of creating it immediately we
        dd it to the list of objects to be created in a single swoop

        :type target: Model to be get_or_create
        :type kwargs: properties to be used to find and create the new object
        """
        created = False
        try:
            instance = target.objects.get(**kwargs)
        except ObjectDoesNotExist:
            instance = target(**kwargs)
            self.instances.add(instance)
            created = True

        return instance, created

    def prepare_save(self, instance: Model):
        """
        Due to the nature of all modifications and such there are some models
        that are in nature get_or_create and not creations
        (we don't want so much additional data)

        This is a recursive function that looks for relationships and
        replaces specific values with their get_or_create counterparts.

        :param instance: model
        :return: instance that is suitable for saving
        """
        from automated_logging.models import (
            Application,
            ModelMirror,
            ModelField,
            ModelEntry,
        )

        if isinstance(instance, Application):
            return Application.objects.get_or_create(name=instance.name)[0]
        elif isinstance(instance, ModelMirror):
            return self.get_or_create(
                ModelMirror,
                name=instance.name,
                application=self.prepare_save(instance.application),
            )[0]
        elif isinstance(instance, ModelField):
            entry, _ = self.get_or_create(
                ModelField,
                name=instance.name,
                mirror=self.prepare_save(instance.mirror),
            )
            if entry.type != instance.type:
                entry.type = instance.type
                # append to instances so that it gets saved with the next .save() call
                self.instances.add(entry)
            return entry

        elif isinstance(instance, ModelEntry):
            entry, _ = self.get_or_create(
                ModelEntry,
                mirror=self.prepare_save(instance.mirror),
                primary_key=instance.primary_key,
            )
            if entry.value != instance.value:
                entry.value = instance.value
                self.instances.add(entry)
            return entry

        # ForeignObjectRel is untouched rn
        for field in [
            f
            for f in instance._meta.get_fields()
            if isinstance(f, ForeignObject)
            and getattr(instance, f.name, None) is not None
        ]:
            setattr(
                instance, field.name, self.prepare_save(getattr(instance, field.name))
            )

        self.instances.add(instance)
        return instance

    def unspecified(self, record: LogRecord) -> None:
        """
        This is for messages that are not sent from django-automated-logging.
        The option to still save these log messages is there. We create
        the event in the handler and then save them.

        :param record:
        :return:
        """
        from automated_logging.models import UnspecifiedEvent, Application
        from automated_logging.signals import unspecified_exclusion
        from django.apps import apps

        event = UnspecifiedEvent()
        if hasattr(record, 'message'):
            event.message = record.message
        event.level = record.levelno
        event.line = record.lineno
        event.file = Path(record.pathname)

        # this is semi-reliable, but I am unsure of a better way to do this.
        applications = apps.app_configs.keys()
        path = Path(record.pathname)
        candidates = [p for p in path.parts if p in applications]
        if candidates:
            # use the last candidate (closest to file)
            event.application = Application(name=candidates[-1])
        elif record.module in applications:
            # if we cannot find the application, we use the module as application
            event.application = Application(name=record.module)
        else:
            # if we cannot determine the application from the application
            # or from the module we presume that the application is unknown
            event.application = Application(name=None)

        if not unspecified_exclusion(event):
            self.prepare_save(event)
            self.save()

    def model(
        self,
        record: LogRecord,
        event: 'ModelEvent',
        modifications: List['ModelValueModification'],
        data: Dict[str, Any],
    ) -> None:
        """
        This is for model specific logging events.
        Compiles the information into an event and saves that event
        and all modifications done.

        :param event:
        :param modifications:
        :param record:
        :param data:
        :return:
        """
        self.prepare_save(event)
        self.save()

        for modification in modifications:
            modification.event = event
            self.prepare_save(modification)
            self.save()

    def m2m(
        self,
        record: LogRecord,
        event: 'ModelEvent',
        relationships: List['ModelRelationshipModification'],
        data: Dict[str, Any],
    ) -> None:
        self.prepare_save(event)
        self.save()

        for relationship in relationships:
            relationship.event = event
            self.prepare_save(relationship)
            self.save()

    def request(self, record: LogRecord, event: 'RequestEvent') -> None:
        """
        The request event already has a model prepared that we just
        need to prepare and save.

        :param record: LogRecord
        :param event: Event supplied via the LogRecord
        :return: nothing
        """

        self.prepare_save(event)
        self.save()

    def emit(self, record: LogRecord) -> None:
        """
        Emit function that gets triggered for every log message in scope.

        The record will be processed according to the action set.
        :param record:
        :return:
        """
        if not hasattr(record, 'action'):
            return self.unspecified(record)

        if record.action == 'model':
            return self.model(record, record.event, record.modifications, record.data)
        elif record.action == 'model[m2m]':
            return self.m2m(record, record.event, record.relationships, record.data)
        elif record.action == 'request':
            return self.request(record, record.event)



"""
Helpers that are used throughout django-automated-logging
"""

from datetime import datetime
from typing import Any, Union, Dict, NamedTuple

from automated_logging.helpers.enums import Operation
from automated_logging.middleware import AutomatedLoggingMiddleware


def namedtuple2dict(root: Union[NamedTuple, Dict]) -> dict:
    """
    transforms nested namedtuple into a dict

    :param root: namedtuple to convert
    :return: dictionary from namedtuple
    """

    output = {}
    if (
        isinstance(root, tuple)
        and hasattr(root, '_serialize')
        and callable(root._serialize)
    ):
        return root._serialize()

    root = root if isinstance(root, dict) else root._asdict()

    def eligible(x):
        """ check if value x is eligible for recursion """
        return isinstance(x, tuple) or isinstance(x, dict)

    for k, v in root.items():
        if isinstance(v, set) or isinstance(v, list):
            output[k] = [namedtuple2dict(i) if eligible(i) else i for i in v]
        else:
            output[k] = namedtuple2dict(v) if eligible(v) else v

    return output


def get_or_create_meta(instance) -> [Any, bool]:
    """
    Simple helper function that creates the dal object
    in _meta.

    :param instance:
    :return:
    """
    return instance, get_or_create_local(instance._meta)


def get_or_create_thread() -> [Any, bool]:
    """
    Get or create the local thread, will always return False as the thread
    won't be created, but the local dal object will.

    get_or_create to conform with the other functions.

    :return: thread, created dal object?
    """
    thread = AutomatedLoggingMiddleware.thread

    return (
        thread,
        get_or_create_local(
            thread,
            {
                'ignore.views': dict,
                'ignore.models': dict,
                'include.views': dict,
                'include.models': dict,
            },
        ),
    )


def get_or_create_local(target: Any, defaults={}, key='dal') -> bool:
    """
    Get or create local storage DAL metadata container,
    where dal specific data is.

    :return: created?
    """

    if not hasattr(target, key):
        setattr(target, key, MetaDataContainer(defaults))
        return True

    return False


def get_or_create_model_event(
    instance, operation: Operation, force=False, extra=False
) -> [Any, bool]:
    """
    Get or create the ModelEvent of an instance.
    This function will also populate the event with the current information.

    :param instance: instance to derive an event from
    :param operation: specified operation that is done
    :param force: force creation of new event?
    :param extra: extra information inserted?
    :return: [event, created?]
    """
    from automated_logging.models import (
        ModelEvent,
        ModelEntry,
        ModelMirror,
        Application,
    )
    from automated_logging.settings import settings

    get_or_create_meta(instance)

    if hasattr(instance._meta.dal, 'event') and not force:
        return instance._meta.dal.event, False

    instance._meta.dal.event = None

    event = ModelEvent()
    event.user = AutomatedLoggingMiddleware.get_current_user()

    if settings.model.snapshot and extra:
        event.snapshot = instance

    if (
        settings.model.performance
        and hasattr(instance._meta.dal, 'performance')
        and extra
    ):
        event.performance = datetime.now() - instance._meta.dal.performance
        instance._meta.dal.performance = None

    event.operation = operation
    event.entry = ModelEntry()
    event.entry.mirror = ModelMirror()
    event.entry.mirror.name = instance.__class__.__name__
    event.entry.mirror.application = Application(name=instance._meta.app_label)
    event.entry.value = repr(instance) or str(instance)
    event.entry.primary_key = instance.pk

    instance._meta.dal.event = event

    return instance._meta.dal.event, True


def function2path(func):
    """ simple helper function to return the module path of a function """
    return f'{func.__module__}.{func.__name__}'


class MetaDataContainer(dict):
    """
    MetaDataContainer is used to store DAL specific metadata
    in various places.

    Values can be retrieved via attribute or key retrieval.

    A dictionary with key attributes can be provided when __init__.
    The key should be the name of the item, the value should be a function
    that gets called when an item with that key does
    not exist gets accessed, to auto-initialize that key.
    """

    def __init__(self, defaults={}):
        super().__init__()

        self.auto = defaults

    def __getitem__(self, item):
        try:
            return super().__getitem__(item)
        except KeyError:
            if item in self.auto:
                self[item] = self.auto[item]()
                return self[item]
            else:
                raise KeyError

    def __getattr__(self, item):
        try:
            return self[item]
        except KeyError:
            raise AttributeError

    def __setattr__(self, key, value):
        self[key] = value



"""
Model definitions for django-automated-logging.
"""
import uuid

from django.conf import settings
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.db.models import (
    CharField,
    ForeignKey,
    CASCADE,
    TextField,
    SmallIntegerField,
    PositiveIntegerField,
    DurationField,
    GenericIPAddressField,
    PositiveSmallIntegerField,
)
from picklefield.fields import PickledObjectField

from automated_logging.helpers import Operation
from automated_logging.helpers.enums import (
    DjangoOperations,
    PastM2MOperationMap,
    ShortOperationMap,
)
from automated_logging.settings import dev


class BaseModel(models.Model):
    """BaseModel that is inherited from every model. Includes basic information."""

    id = models.UUIDField(default=uuid.uuid4, primary_key=True, db_index=True)

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True


class Application(BaseModel):
    """
    Used to save from which application an event or model originates.
    This is used to group by application.

    The application name can be null,
    if the name is None, then the application is unknown.
    """

    name = CharField(max_length=255, null=True)

    class Meta:
        verbose_name = "Application"
        verbose_name_plural = "Applications"

    class LoggingIgnore:
        complete = True

    def __str__(self):
        return self.name or "Unknown"


class ModelMirror(BaseModel):
    """
    Used to mirror properties of models - this is used to preserve logs of
    models removed to make the logs independent of the presence of the model
    in the application.
    """

    name = CharField(max_length=255)
    application = ForeignKey(Application, on_delete=CASCADE)

    class Meta:
        verbose_name = "Model Mirror"
        verbose_name_plural = "Model Mirrors"

    class LoggingIgnore:
        complete = True

    def __str__(self):
        return self.name


class ModelField(BaseModel):
    """
    Used to mirror properties of model fields - this is used to preserve logs of
    models and fields that might be removed/modified and have them independent
    of the actual field.
    """

    name = CharField(max_length=255)

    mirror = ForeignKey(ModelMirror, on_delete=CASCADE)
    type = CharField(max_length=255)  # string of type

    class Meta:
        verbose_name = "Model Field"
        verbose_name_plural = "Model Fields"

    class LoggingIgnore:
        complete = True


class ModelEntry(BaseModel):
    """
    Used to mirror the evaluated model value (via repr) and primary key and
    to ensure the log integrity independent of presence of the entry.
    """

    mirror = ForeignKey(ModelMirror, on_delete=CASCADE)

    value = TextField()  # (repr)
    primary_key = TextField()

    class Meta:
        verbose_name = "Model Entry"
        verbose_name_plural = "Model Entries"

    class LoggingIgnore:
        complete = True

    def __str__(self) -> str:
        return f'{self.mirror.name}' f'(pk="{self.primary_key}", value="{self.value}")'

    def long(self) -> str:
        """
        long representation
        """

        return f'{self.mirror.application.name}.{self})'

    def short(self) -> str:
        """
        short representation
        """
        return f'{self.mirror.name}({self.primary_key})'


class ModelEvent(BaseModel):
    """
    Used to record model entry events, like modification, removal or adding of
    values or relationships.
    """

    operation = SmallIntegerField(
        validators=[MinValueValidator(-1), MaxValueValidator(1)],
        null=True,
        choices=DjangoOperations,
    )

    user = ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=CASCADE, null=True
    )  # maybe don't cascade?
    entry = ForeignKey(ModelEntry, on_delete=CASCADE)

    # modifications = None  # One2Many -> ModelModification
    # relationships = None  # One2Many -> ModelRelationship

    # v experimental, opt-in (pickled object)
    snapshot = PickledObjectField(null=True)
    performance = DurationField(null=True)

    class Meta:
        verbose_name = "Model Event"
        verbose_name_plural = "Model Events"

    class LoggingIgnore:
        complete = True


class ModelValueModification(BaseModel):
    """
    Used to record the model entry event modifications of simple values.

    The operation attribute can have 4 valid values:
    -1 (delete), 0 (modify), 1 (create), None (n/a)

    previous and current record the value change that happened.
    """

    operation = SmallIntegerField(
        validators=[MinValueValidator(-1), MaxValueValidator(1)],
        null=True,
        choices=DjangoOperations,
    )

    field = ForeignKey(ModelField, on_delete=CASCADE)

    previous = TextField(null=True)
    current = TextField(null=True)

    event = ForeignKey(ModelEvent, on_delete=CASCADE, related_name='modifications')

    class Meta:
        verbose_name = "Model Entry Event Value Modification"
        verbose_name_plural = "Model Entry Event Value Modifications"

    class LoggingIgnore:
        complete = True

    def __str__(self) -> str:
        return (
            f'[{self.field.mirror.application.name}:'
            f'{self.field.mirror.name}:'
            f'{self.field.name}] '
            f'{self.previous} -> {self.current}'
        )

    def short(self) -> str:
        """
        short representation analogue of __str__
        """
        operation = Operation(self.operation)
        shorthand = {v: k for k, v in ShortOperationMap.items()}[operation]

        return f'{shorthand}{self.field.name}'


class ModelRelationshipModification(BaseModel):
    """
    Used to record the model entry even modifications of relationships. (M2M, Foreign)


    The operation attribute can have 4 valid values:
    -1 (delete), 0 (modify), 1 (create), None (n/a)

    field is the field where the relationship changed (entry got added or removed)
    and model is the entry that got removed/added from the relationship.
    """

    operation = SmallIntegerField(
        validators=[MinValueValidator(-1), MaxValueValidator(1)],
        null=True,
        choices=DjangoOperations,
    )

    field = ForeignKey(ModelField, on_delete=CASCADE)
    entry = ForeignKey(ModelEntry, on_delete=CASCADE)

    event = ForeignKey(ModelEvent, on_delete=CASCADE, related_name='relationships')

    class Meta:
        verbose_name = "Model Entry Event Relationship Modification"
        verbose_name_plural = "Model Entry Event Relationship Modifications"

    class LoggingIgnore:
        complete = True

    def __str__(self) -> str:
        operation = Operation(self.operation)
        past = {v: k for k, v in PastM2MOperationMap.items()}[operation]

        return (
            f'[{self.field.mirror.application}:'
            f'{self.field.mirror.name}:'
            f'{self.field.name}] '
            f'{past} {self.entry}'
        )

    def short(self) -> str:
        """
        short representation
        """
        operation = Operation(self.operation)
        shorthand = {v: k for k, v in ShortOperationMap.items()}[operation]
        return f'{shorthand}{self.entry.short()}'

    def medium(self) -> [str, str]:
        """
        short representation analogue of __str__ with additional field context
        :return:
        """
        operation = Operation(self.operation)
        shorthand = {v: k for k, v in ShortOperationMap.items()}[operation]

        return f'{shorthand}{self.field.name}', f'{self.entry.short()}'


class RequestContext(BaseModel):
    """
    Used to record contents of request and responses and their type.
    """

    content = PickledObjectField(null=True)
    type = CharField(max_length=255)

    class LoggingIgnore:
        complete = True


class RequestEvent(BaseModel):
    """
    Used to record events of requests that happened.

    uri is the accessed path and data is the data that was being transmitted
    and is opt-in for collection.

    status and method are their respective HTTP equivalents.
    """

    user = ForeignKey(settings.AUTH_USER_MODEL, on_delete=CASCADE, null=True)

    # to mitigate "max_length"
    uri = TextField()

    request = ForeignKey(
        RequestContext, on_delete=CASCADE, null=True, related_name='request_context'
    )
    response = ForeignKey(
        RequestContext, on_delete=CASCADE, null=True, related_name='response_context'
    )

    status = PositiveSmallIntegerField()
    method = CharField(max_length=32)

    application = ForeignKey(Application, on_delete=CASCADE)

    ip = GenericIPAddressField(null=True)

    class Meta:
        verbose_name = "Request Event"
        verbose_name_plural = "Request Events"

    class LoggingIgnore:
        complete = True


class UnspecifiedEvent(BaseModel):
    """
    Used to record unspecified internal events that are dispatched via
    the python logging library. saves the message, level, line, file and application.
    """

    message = TextField(null=True)
    level = PositiveIntegerField(default=20)

    line = PositiveIntegerField(null=True)
    file = TextField(null=True)

    application = ForeignKey(Application, on_delete=CASCADE)

    class Meta:
        verbose_name = "Unspecified Event"
        verbose_name_plural = "Unspecified Events"

    class LoggingIgnore:
        complete = True


if dev:
    # if in development mode (set when testing or development)
    # import extra models
    from automated_logging.tests.models import *
